ARRAY = 'array'
AUTHOR = 'author'
BEHAVIOR = 'behavior'
BOOLEAN = 'boolean'
CELL = 'cell'
CELL_REF = 'cellref'
CELL_TYPE = 'celltype'
COMMENT = 'comment'
CONTENTS = 'contents'
DATA_ORIGIN = 'dataorigin'
DESIGN = 'design'
DIRECTION = 'direction'
DOCUMENT = 'document'
E = 'e'
EDIF = 'edif'
EDIF_VERSION = 'edifversion'
EDIF_LEVEL = 'ediflevel'
EXTERNAL = 'external'
FALSE = 'false'
GENERIC = 'generic'
GRAPHIC = 'graphic'
INOUT = 'inout'
INPUT = 'input'
INSTANCE = 'instance'
INSTANCE_REF = 'instanceref'
INTEGER = 'integer'
INTERFACE = 'interface'
KEYWORD_MAP = 'keywordmap'
KEYWORD_LEVEL = 'keywordlevel'
LEFT_PAREN = '('
LIBRARY = 'library'
LIBRARY_REF = 'libraryref'
LOGICMODEL = 'logicmodel'
MASKLAYOUT = 'masklayout'
MEMBER = 'member'
METAX = 'metax'
NET = 'net'
NETLIST = 'netlist'
OUTPUT = 'output'
OWNER = 'owner'
PCBLAYOUT = 'netlist'
PORT = 'port'
PORT_REF = 'portref'
PROGRAM = 'program'
PROPERTY = 'property'
NUMBER = 'number'
NUMBER_DEFINITION = 'numberdefinition'
RENAME = 'rename'
RIGHT_PAREN = ')'
RIPPER = 'ripper'
SCHEMATIC = 'schematic'
STATUS = 'status'
STRANGER = 'stranger'
STRING = 'string'
SYMBOLIC = 'symbolic'
TECHNOLOGY = 'technology'
TIE = 'tie'
TIME_STAMP = 'timestamp'
TRUE = 'true'
UNIT = 'unit'
USER_DATA = 'userdata'
VERSION = 'version'
VIEW = 'view'
VIEW_MAP = 'viewmap'
VIEW_REF = 'viewref'
VIEW_TYPE = 'viewtype'
WRITTEN = 'written'

ARRAY_RELATED_INFO = 'arrayrelatedinfo'
DESIGNATOR = 'designator'
PORT_BUNDLE = 'portbundle'
PROTECTION_FRAME = 'protectionframe'
SYMBOL = 'symbol'
PARAMETER = 'parameter'
JOINED = 'joined'
MUST_JOIN = 'mustjoin'
WEAK_JOINED = 'weakjoined'
PERMUTABLE = 'permutable'
TIMING = 'timing'
SIMULATE = 'simulate'
WEAK_JOINED = 'weakjoined'

UNUSED = 'unused'
DC_FANIN_LOAD = 'dc_fanin_load'
DC_FANOUT_LOAD = 'dc_fanout_load'
DC_MAX_FANIN = 'dc_max_fanout'
DC_MAX_FANOUT = 'dc_max_fanin'
AC_LOAD = 'ac_load'
PORT_DELAY = 'port_delay'

OFF_PAGE_CONNECTOR = 'offpageconnector'
FIGURE = 'figure'
SECTION = 'section'
NET_BUNDLE = 'netbundle'
PAGE = 'page'
COMMENT_GRAPHICS = 'commentgraphics'
PORT_IMPLEMENTATION = 'portimplementation'
WHEN = 'when'
FOLLOW = 'follow'
LOGIC_PORT = 'logicport'
BOUNDING_BOX = 'boundingbox'

VIEW_LIST = 'viewlist'

PORT_LIST = 'portlist'
GLOBAL_PORT_REF = 'globalportref'

POINT = 'point'
MI_NO_MAX = 'minomax'


from spydrnet.ir.element import Element
from spydrnet.ir.port import Port
from spydrnet.ir.cable import Cable
from spydrnet.ir.instance import Instance
from spydrnet.ir.outerpin import OuterPin
from spydrnet.ir.views.listview import ListView
from spydrnet.ir.views.setview import SetView
from spydrnet.global_state import global_callback


class Definition(Element):
    """
    Represents a definition of a cell, module, entity/architecture, or paralleled structure object.

    Contains a pointer to parent library, ports, cables, and instances.
    """
    __slots__ = ['_library', '_ports', '_cables', '_children', '_references']

    def __init__(self):
        super().__init__()
        self._library = None
        self._ports = list()
        self._cables = list()
        self._children = list()
        self._references = set()

    @property
    def library(self):
        """
        Get the library that contains this definition
        """
        return self._library

    @property
    def ports(self):
        """
        get the ports that are instanced in this definition
        """
        return ListView(self._ports)

    @ports.setter
    def ports(self, value):
        """
        Reorder ports that are instanced in this definition. Use remove_port and add_port to remove and add ports
        respectively

        parameters
        ----------

        value - (List of type Port objects) the reordered list of ports
        """
        target = list(value)
        target_set = set(target)
        assert len(target) == len(target_set) and set(self._ports) == target_set, \
            "Set of values do not match, this function can only reorder values, values must be unique"
        self._ports = target

    @property
    def cables(self):
        """
        get the cables that are instanced in this definition
        """
        return ListView(self._cables)

    @cables.setter
    def cables(self, value):
        """
        Reorder the cables in this definition. Use add_cable and remove_cable to add or remove cables.


        parameters
        ----------

        value - (List of type Cable objects) the reordered list of cables
        """
        target = list(value)
        target_set = set(target)
        assert len(target) == len(target_set) and set(self._cables) == set(target), \
            "Set of values do not match, this function can only reorder values, values must be unique"
        self._cables = target

    @property
    def children(self):
        """
        return a list of all instances instantiated in this definition
        """
        return ListView(self._children)

    @children.setter
    def children(self, value):
        """
        reorder the list of instances instantiated in this definition use add_child and remove_child to add or remove
        instances to or from the definition

        parameters
        ----------

        value - (List of type Instance objects) the reordered list of instances
        """
        target = list(value)
        target_set = set(target)
        assert len(target) == len(target_set) and set(self._children) == target_set, \
            "Set of values do not match, this function can only reorder values, values must be unique"
        self._children = target

    @property
    def references(self):
        """
        get a list of all the instances of this definition
        """
        return SetView(self._references)

    def is_leaf(self):
        """
        check to see if this definition represents a leaf cell. Leaf cells are cells with no children instances or no
        children cables. Blackbox cells are considered leaf cells as well as direct pass through cells with cables only
        """
        if len(self._children) > 0 or len(self._cables) > 0:
            return False
        return True

    def create_port(self):
        """
        create a port, add it to the definition, and return that port
        """
        port = Port()
        self.add_port(port)
        return port

    def add_port(self, port, position=None):
        """
        add a preexisting port to the definition. this port must not be a member of any definition

        parameters
        ----------

        port - (Port) the port to add to the definition

        position - (int, default None) the index in the port list at which to add the port
        """
        assert port.definition is not self, "Port already included in definition"
        assert port.definition is None, "Port already belongs to a different definition"
        global_callback._call_definition_add_port(self, port)
        if position is not None:
            self._ports.insert(position, port)
        else:
            self._ports.append(port)
        port._definition = self
        for reference in self.references:
            for pin in port.pins:
                reference._pins[pin] = OuterPin(reference, pin)

    def remove_port(self, port):
        """
        remove a port from the definition. This port must be a member of the definition in order to be removed

        parameters
        ----------

        port - (Port) the port to be removed
        """
        assert port.definition == self, "Port is not included in definition"
        self._remove_port(port)
        self._ports.remove(port)

    def remove_ports_from(self, ports):
        """
        remove a set of ports from the definition. All these ports must be included in the definition

        parameters
        ----------

        ports - (Set containing Port type objects) the ports to remove from the definition
        """
        if isinstance(ports, set):
            excluded_ports = ports
        else:
            excluded_ports = set(ports)
        assert all(x.definition == self for x in excluded_ports), "Some ports to remove are not included in the " \
                                                                  "definition."
        for port in excluded_ports:
            self._remove_port(port)
        self._ports = list(x for x in self._ports if x not in excluded_ports)

    def _remove_port(self, port):
        """
        internal function to dissociate the port from the definition

        Parameters
        ----------

        port - (Port) the port to remove from the definition
        """
        global_callback._call_definition_remove_port(self, port)
        for reference in self.references:
            for pin in port.pins:
                outer_pin = reference.pins[pin]
                wire = outer_pin.wire
                if wire:
                    wire.disconnect_pin(outer_pin)
                del reference._pins[pin]
                outer_pin._instance = None
                outer_pin._inner_pin = None
        port._definition = None

    def create_child(self):
        """
        create an instance to add to the definition, add it, and return the instance.
        """
        instance = Instance()
        self.add_child(instance)
        return instance

    def add_child(self, instance, position=None):
        """
        Add an existing instance to the definition. This instance must not already be included in a definition

        parameters
        ----------

        instance - (Instance) the instance to add as a child of the definition

        position - (int, default None) the index in the children list at which to add the instance.
        """
        assert instance.parent is not self, "Instance already included in definition"
        assert instance.parent is None, "Instance already belongs to a different definition"
        global_callback._call_definition_add_child(self, instance)
        if position is not None:
            self._children.insert(position, instance)
        else:
            self._children.append(instance)
        instance._parent = self

    def remove_child(self, child):
        """
        remove an instance from the definition. The instance must be a member of the definition already

        parameters
        ----------

        instance - (Instance) the instance to be removed from the definition
        """
        assert child.parent == self, "Instance is not included in definition"
        self._remove_child(child)
        self._children.remove(child)

    def remove_children_from(self, children):
        """
        remove a set of instances from the definition. All instances must be members of the definition

        parameters
        ----------

        children - (Set of Instance type objects) the children to be removed from the definition
        """
        if isinstance(children, set):
            excluded_children = children
        else:
            excluded_children = set(children)
        assert all(x.parent == self for x in excluded_children), "Some children are not parented by the definition"
        included_children = list()
        for child in self._children:
            if child not in excluded_children:
                included_children.append(child)
            else:
                self._remove_child(child)
        self._children = included_children

    def _remove_child(self, child):
        """
        internal function for dissociating a child instance from the definition.
        """
        global_callback._call_definition_remove_child(self, child)
        child._parent = None

    def create_cable(self):
        """
        create a cable, add it to the definition, and return the cable.
        """
        cable = Cable()
        self.add_cable(cable)
        return cable

    def add_cable(self, cable, position=None):
        """
        add a cable to the definition. The cable must not already be a member of another definition.

        parameters
        ----------

        cable - (Cable) the cable to be added

        position - (int, default None) the position in the cable list at which to add the cable
        """
        assert cable.definition is not self, "Cable already included in definition"
        assert cable.definition is None, "Cable already belongs to a different definition"
        global_callback._call_definition_add_cable(self, cable)
        if position is not None:
            self._cables.insert(position, cable)
        else:
            self._cables.append(cable)
        cable._definition = self

    def remove_cable(self, cable):
        """
        remove a cable from the definition. The cable must be a member of the definition.

        parameters
        ----------

        cable - (Cable) the cable to be removed from the definition
        """
        assert cable.definition == self, "Cable is not included in definition"
        self._remove_cable(cable)
        self._cables.remove(cable)

    def remove_cables_from(self, cables):
        """
        remove a set of cables from the definition. The cables must be members of the definition

        parameters
        ----------

        cables - (Set of Cable type objects) the cables to be remove from the definition
        """
        if isinstance(cables, set):
            excluded_cables = cables
        else:
            excluded_cables = set(cables)
        assert all(x.definition == self for x in excluded_cables), "Some cables are not included in the definition"
        included_cables = list()
        for cable in self._cables:
            if cable not in excluded_cables:
                included_cables.append(cable)
            else:
                self._remove_cable(cable)
        self._cables = included_cables

    def _remove_cable(self, cable):
        """
        dissociate the cable from this definition. This function is internal and should not be called.
        """
        global_callback._call_definition_remove_cable(self, cable)
        cable._definition = None



from functools import partial
import re
import zipfile
import io
import os

class EdifTokenizer:
    @staticmethod
    def from_stream(stream):
        tokenizer = EdifTokenizer(stream)
        return tokenizer
    
    @staticmethod
    def from_string(string):
        string_stream = io.StringIO(string)
        tokenizer = EdifTokenizer(string_stream)
        return tokenizer

    @staticmethod
    def from_filename(filename):
        tokenizer = EdifTokenizer(filename)
        return tokenizer
    
    def __init__(self, input_source):
        self.token = None
        self.next_token = None
        self.line_number = 0

        if isinstance(input_source, str):
            if zipfile.is_zipfile(input_source):
                zip = zipfile.ZipFile(input_source)
                filename = os.path.basename(input_source)
                filename = filename[:filename.rindex(".")]
                stream = zip.open(filename)
                stream = io.TextIOWrapper(stream)
                self.input_stream = stream
            else:
                self.input_stream = open(input_source, 'r')
        else:
            if isinstance(input_source, io.TextIOBase) is False:
                self.input_stream = io.TextIOWrapper(input_source)
            else:
                self.input_stream = input_source

        self.generator = self.generate_tokens()

    def __del__(self):
        if hasattr(self, "input_stream"):
            self.close()

    def has_next(self):
        try:
            self.peek()
            return True
        except StopIteration:
            return False

    def next(self):
        if self.next_token:
            self.token = self.next_token
            self.next_token = None
        else:
            self.token = next(self.generator)
        return self.token

    def peek(self):
        if self.next_token:
            return self.next_token
        else:
            self.next_token = next(self.generator)
            return self.next_token

    def generate_tokens(self):
        try:
            self.line_number = 1
            in_quote = False
            token_buffer = list()
            for buffer in iter(partial(self.input_stream.read, 32768), ""):
                for ch in buffer:
                    if ch == '\n': self.line_number += 1
                    if in_quote:
                        if ch in {"\n", "\r"}:
                            continue
                        token_buffer.append(ch)
                        if ch == '"':
                            in_quote = False
                            token = ''.join(token_buffer)
                            token_buffer.clear()
                            yield token
                    elif ch == '"':
                        in_quote = True
                        token_buffer.append(ch)
                    elif ch in {'(', ')'}:
                        if token_buffer:
                            token = ''.join(token_buffer)
                            token_buffer.clear()
                            yield token
                        yield ch
                    elif ch in {'\r', '\n', '\t', ' '}:
                        if token_buffer:
                            token = ''.join(token_buffer)
                            token_buffer.clear()
                            yield token
                    else:
                        token_buffer.append(ch)

            if token_buffer:
                    token = ''.join(token_buffer)
                    token_buffer.clear()
                    yield token
        finally:
            self.input_stream.close()

    def close(self):
        if self.input_stream:
            self.input_stream.close()

    def expect(self, other):
        if not self.token_equals(other):
            raise RuntimeError("Parse error: Expecting {} on line {}, recieved {}".format(other, self.line_number, self.token))

    def peek_equals(self, other):
        peek_token = self.peek()
        return self.equals(peek_token, other)
    
    def token_equals(self, other):
        return self.equals(self.token, other)
    
    @staticmethod
    def equals(this, that):
        if this == that:
            return True
        else:
            lowercase_this = this.lower()
            if lowercase_this == that:
                return True
            elif lowercase_this == that.lower():
                return True
        return False

    def expect_valid_identifier(self):
        if self.is_valid_identifier() is False:
            raise RuntimeError("Parse error: Expecting EDIF identifier on line {}, recieved {}".format(self.line_number, self.token))

    def is_valid_identifier(self):
        if re.match(r"[a-zA-Z]|&\a*", self.token) and len(self.token) <= 256:
            return True
        return False

    def expect_valid_integerToken(self):
        if self.is_valid_integerToken() is False:
            raise RuntimeError("Parse error: Expecting integerToken on line {}, recieved {}".format(self.line_number, self.token))

    def is_valid_integerToken(self):
        if re.match(r"[-+]?\d+", self.token):
            return True
        return False

    def expect_valid_stringToken(self):
        if self.is_valid_stringToken() is False:
            raise RuntimeError("Parse error: Expecting stringToken on line {}, recieved {}".format(self.line_number, self.token))

    def is_valid_stringToken(self):
        if re.match(r'"(?:[a-zA-Z]|(?:%[ \t\n\r]*(?:(?:[-+]?\d+[ \t\n\r]+)*(?:[-+]?\d+))*[ \t\n\r]*%)|[0-9]|[\!\#\$\&\'\(\)\*\+\,\-\.\/\:\;\<\=\>\?\@\[\\\]\^\_\`\{\|\}\~]|[ \t\n\r])*"', self.token):
            return True
        return False

if __name__ == "__main__":
    # filename = r"C:\Users\keller\workplace\SpyDrNet\data\large_edif\osfbm.edf"
    import cProfile
    def run():
        filename = r"C:\Users\akeller9\workspace\SpyDrNet\data\large_edif\osfbm.edf"
        tokenizer = EdifTokenizer.from_filename(filename)
        count = 0
        for token in tokenizer.generator:
            if count < 100:
                print(token)
            count += 1
        print(count)
    cProfile.run("run()")


import unittest
from spydrnet.ir.views.dictview import DictView

class TestDictView(unittest.TestCase):
    def setUp(self) -> None:
        self.dict_1 = dict(map(lambda x: (x, str(x)), range(10)))
        self.dict_2 = dict(map(lambda x: (x, str(x)), range(10, 20)))
        self.dict_view_1 = DictView(self.dict_1)
        self.dict_view_2 = DictView(self.dict_2)
        self.dict_3 = dict(map(lambda x: (x, str(x)), range(30, 30)))

    def test_eq(self):
        self.assertEqual(self.dict_view_1, self.dict_1)
        self.assertEqual(self.dict_1, self.dict_view_1)
        self.assertEqual(DictView(self.dict_1), self.dict_view_1)
        self.assertFalse(DictView(self.dict_2) == self.dict_view_1)

    def test_hash(self):
        self.assertRaises(TypeError, hash, self.dict_view_1)

    def test_len(self):
        self.assertEqual(len(self.dict_1), len(self.dict_view_1))

    def test_ne(self):
        self.assertNotEqual(self.dict_view_2, self.dict_1)
        self.assertNotEqual(self.dict_2, self.dict_view_1)
        self.assertNotEqual(DictView(self.dict_2), self.dict_view_1)
        self.assertTrue(DictView(self.dict_2) != self.dict_view_1)

    def test_repr(self):
        self.assertEqual(repr(self.dict_view_1), repr(self.dict_1))

    def test_str(self):
        self.assertEqual(str(self.dict_view_1), str(self.dict_1))

    def test_contains(self):
        self.assertTrue(5 in self.dict_view_1)
        self.assertFalse(10 in self.dict_view_1)

    def test_getitem(self):
        for ii in range(10):
            self.assertTrue(self.dict_view_1[ii] == str(ii))

    def test_iter(self):
        for ii, jj in zip(self.dict_view_1, range(10)):
            self.assertEqual(ii, jj)

    def test_get(self):
        for ii in range(11):
            self.assertTrue(self.dict_view_1.get(ii, '10') == str(ii))

    def test_copy(self):
        self.assertEqual(dict(map(lambda x: (x, str(x)), range(10))), self.dict_view_1.copy())

    def test_fromkeys(self):
        self.assertEqual(self.dict_view_1.fromkeys(range(10), 5), dict(map(lambda x: (x, 5), range(10))))

    def test_keys(self):
        for ii, jj in zip(self.dict_view_1.keys(), range(10)):
            self.assertEqual(ii, jj)

    def test_values(self):
        for ii, jj in zip(self.dict_view_1.values(), range(10)):
            self.assertEqual(ii, str(jj))

    def test_items(self):
        for ii, jj in self.dict_view_1.items():
            self.assertEqual(str(ii), jj)



import unittest

import spydrnet as sdn
from spydrnet.ir.element import Element


class TestInstance(unittest.TestCase):
    def setUp(self) -> None:
        self.instance = sdn.Instance()

    def test_constructor(self):
        self.assertIsInstance(self.instance, Element, "Instance should extend element")
        instance2 = sdn.Instance()
        self.assertNotEqual(self.instance, instance2, "Unique objects are considered equal")

    def test_reference_assignment(self):
        definition = sdn.Definition()
        self.instance.reference = definition
        self.assertEqual(self.instance.reference, definition)
        self.assertTrue(self.instance in definition.references)

    def test_reference_assignment_with_pins(self):
        definition = sdn.Definition()
        port = definition.create_port()
        pin1 = port.create_pin()
        pin2 = port.create_pin()

        self.instance.reference = definition
        self.assertTrue(pin1 in self.instance.pins)
        self.assertTrue(pin2 in self.instance.pins)
        outer_pin1 = self.instance.pins[pin1]
        outer_pin2 = self.instance.pins[pin2]
        self.assertIsInstance(outer_pin1, sdn.OuterPin)
        self.assertIsInstance(outer_pin2, sdn.OuterPin)
        self.assertEqual(outer_pin1.instance, self.instance)
        self.assertEqual(outer_pin2.instance, self.instance)
        self.assertEqual(outer_pin1.inner_pin, pin1)
        self.assertEqual(outer_pin2.inner_pin, pin2)

        wire = sdn.Wire()
        wire.connect_pin(outer_pin1)
        wire.connect_pin(outer_pin2)
        self.instance.reference = None
        self.assertEqual(len(self.instance.pins), 0)
        self.assertIsNone(outer_pin1.wire)
        self.assertIsNone(outer_pin2.wire)
        self.assertIsNone(outer_pin1.instance)
        self.assertIsNone(outer_pin2.instance)
        self.assertIsNone(outer_pin2.inner_pin)
        self.assertIsNone(outer_pin2.inner_pin)

    def test_post_assignment_pin_and_port_removal(self):
        definition = sdn.Definition()
        port = definition.create_port()
        pin1 = port.create_pin()
        pin2 = port.create_pin()

        self.instance.reference = definition
        outer_pin1 = self.instance.pins[pin1]
        outer_pin2 = self.instance.pins[pin2]

        wire = sdn.Wire()
        wire.connect_pin(outer_pin1)
        wire.connect_pin(outer_pin2)

        port.remove_pin(pin1)
        definition.remove_port(port)

        self.assertIsNone(outer_pin1.wire)
        self.assertIsNone(outer_pin1.instance)
        self.assertIsNone(outer_pin1.inner_pin)
        self.assertIsNone(outer_pin2.wire)
        self.assertIsNone(outer_pin2.instance)
        self.assertIsNone(outer_pin2.inner_pin)
        self.assertFalse(outer_pin1 in wire.pins)
        self.assertFalse(outer_pin2 in wire.pins)
        self.assertFalse(outer_pin1 in self.instance.pins)
        self.assertFalse(outer_pin2 in self.instance.pins)
        self.assertFalse(pin1 in self.instance.pins)
        self.assertFalse(pin2 in self.instance.pins)

    def test_post_assignment_pin_and_port_creation(self):
        definition = sdn.Definition()
        self.instance.reference = definition
        port = definition.create_port()
        pin1 = port.create_pin()
        pin2 = port.create_pin()
        port2 = sdn.Port()
        pin3 = port2.create_pin()
        definition.add_port(port2)
        pin4 = port2.create_pin()

        outer_pin1 = self.instance.pins[pin1]
        outer_pin2 = self.instance.pins[pin2]
        outer_pin3 = self.instance.pins[pin3]
        outer_pin4 = self.instance.pins[pin4]

        wire = sdn.Wire()
        wire.connect_pin(outer_pin1)
        wire.connect_pin(outer_pin2)
        wire.connect_pin(outer_pin3)
        wire.connect_pin(outer_pin4)
        inner_pins = [pin1, pin2, pin3, pin4]
        outer_pins = [outer_pin1, outer_pin2, outer_pin3, outer_pin4]
        for outer_pin, inner_pin in zip(outer_pins, inner_pins):
            self.assertEqual(outer_pin.wire, wire)
            self.assertTrue(outer_pin in wire.pins)
            self.assertEqual(outer_pin.instance, self.instance)
            self.assertEqual(outer_pin.inner_pin, inner_pin)
            self.assertTrue(outer_pin in self.instance.pins)
            self.assertTrue(inner_pin in self.instance.pins)

    def test_reference_reassignment(self):
        definition = sdn.Definition()
        port = definition.create_port()
        pin1 = port.create_pin()
        pin2 = port.create_pin()
        self.instance.reference = definition
        self.assertTrue(self.instance in definition.references)
        outer_pin1 = self.instance.pins[pin1]
        outer_pin2 = self.instance.pins[pin2]

        definition2 = sdn.Definition()
        port = definition2.create_port()
        pin1 = port.create_pin()
        pin2 = port.create_pin()
        self.instance.reference = definition2
        self.assertTrue(self.instance not in definition.references)
        self.assertTrue(self.instance in definition2.references)
        self.assertEqual(outer_pin1, self.instance.pins[pin1])
        self.assertEqual(outer_pin2, self.instance.pins[pin2])

    def test_reference_removal(self):
        definition = sdn.Definition()
        port = definition.create_port()
        pin1 = port.create_pin()
        pin2 = port.create_pin()
        self.instance.reference = definition
        self.assertTrue(self.instance in definition.references)
        outer_pin1 = self.instance.pins[pin1]
        outer_pin2 = self.instance.pins[pin2]
        self.assertEqual(outer_pin1.inner_pin, pin1)
        self.assertEqual(outer_pin2.inner_pin, pin2)

        del self.instance.reference
        self.assertTrue(self.instance not in definition.references)
        self.assertIsNone(outer_pin1.inner_pin)
        self.assertIsNone(outer_pin2.inner_pin)




import unittest
import os
import tempfile
import glob
import shutil

from spydrnet.compare.compare_netlists import Comparer
import spydrnet as sdn


class TestCompareNetlists(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.dir_of_edif_netlists = os.path.join(sdn.base_dir, "support_files", "EDIF_netlists")
        cls.edif_files = sorted(glob.glob(os.path.join(cls.dir_of_edif_netlists, "*.edf.zip")), key=os.path.getsize)

    @unittest.skip("Test takes a long time right now.")
    def test_large_edif(self):
        for ii, filename in enumerate(self.edif_files):
            if os.path.getsize(filename) <= 1024 * 10:
                continue
            self.compare_parser_and_composer_on(filename, ii)

    def test_small_edif(self):
        for ii, filename in enumerate(self.edif_files):
            if os.path.getsize(filename) > 1024 * 10:
                continue
            self.compare_parser_and_composer_on(filename, ii)

    def compare_parser_and_composer_on(self, filename, ii):
            with self.subTest(i=ii):
                if os.path.exists("temp"):
                    shutil.rmtree("temp")
                print(filename)
                orig_netlist = sdn.parse(filename)
                with tempfile.TemporaryDirectory() as tempdirname:
                    try:
                        basename_without_final_ext = os.path.splitext(os.path.basename(filename))[0]
                        composer_filename = os.path.join(tempdirname, basename_without_final_ext)
                        sdn.compose(composer_filename, orig_netlist)
                        composer_netlist = sdn.parse(composer_filename)
                    except Exception as e:
                        shutil.copytree(tempdirname, "temp")
                        raise e
                comparer = Comparer(orig_netlist, composer_netlist)
                comparer.run()


if __name__ == '__main__':
    unittest.main()



from spydrnet.ir.element import Element

class Bundle(Element):
    """
    parent class of ports and cables. Since both of these objects represent arrays of objects they both inherit from
    this parent class.
    """
    __slots__ = ['_definition', '_is_downto', '_is_scalar', '_lower_index']

    def __init__(self):
        super().__init__()
        self._definition = None
        self._is_downto = True
        self._is_scalar = True
        self._lower_index = 0

    @property
    def definition(self):
        """
        Get the definition that this bundle belongs to. The definition is responsible for changing this value.
        """
        return self._definition

    @property
    def is_downto(self):
        """
        get the downto status of the bundle. Downto is False if the right index is higher than the left one. True
        otherwise
        """
        return self._is_downto

    @is_downto.setter
    def is_downto(self, value):
        """
        change the downto value Downto is False if the right index is higher than the left index. True otherwise.

        parameters
        ----------

        value - (boolean) True if the value is downto False if the value is to.
        """
        self._is_downto = value

    def _items(self):
        """
        this function must be overridden in classes which extend this to return either a list of pins or wires
        """
        return None

    @property
    def is_scalar(self):
        """
        Return True if the item is a scalar False otherwise. the item is not a scalar if it has more than one pin or
        wire in it. if it has one pin or wire in it it may be a scalar. This mimics vhdl's downto usage which can
        represent single pin arrays ie. std_logic_vector(0 downto 0) which would have a single pin but not be a scalar.
        """
        _items = self._items()
        if _items and len(_items) > 1:
            return False
        return self._is_scalar

    @is_scalar.setter
    def is_scalar(self, value):
        """
        set the scalar status of single item bundles.
        The item is not a scalar if it has more than one pin or wire in it. if it has one or zero pins this function
        can be used to set the value or wire in it it may be a scalar. This mimics vhdl's downto usage which can
        represent single pin arrays ie. std_logic_vector(0 downto 0) which would have a single pin but not be a scalar.

        parameters
        ----------

        value - (boolean) True if the item is to be a scalar False if it is not. Multi element bundles cannot set
        is_scalar to True
        """
        _items = self._items()
        if _items and len(_items) > 1 and value is True:
            raise RuntimeError("Cannot set is_scalar to True on a multi-item bundle")
        else:
            self._is_scalar = value

    @property
    def is_array(self):
        """
        this is the logical inverse of is_scalar. see the is_scalar documentation for more insight into the properties
        of this value
        """
        return not self.is_scalar

    @is_array.setter
    def is_array(self, value):
        """
        this is the logical inverse of is_scalar. see the is_scalar documentation for more insight into the properties
        of this value

        parameters
        ----------

        value - (boolean) True if the object is an array. False otherwise. Multi element bundles cannot set is_array to
        false.
        """
        _items = self._items()
        if _items and len(_items) > 1 and value is False:
            raise RuntimeError("Cannot set is_array to False on a multi-item bundle")
        else:
            self._is_scalar = not value

    @property
    def lower_index(self):
        """
        get the value of the lower index of the array. this would be the right index in the case of downto and the left
        in the case of to
        """
        return self._lower_index

    @lower_index.setter
    def lower_index(self, value):
        """
        set the lower index of the array. in the case of to this is the left index and the right in the case of downto

        parameters
        ----------

        value - (int) the lower index value for the bundle.
        """
        self._lower_index = value



import unittest

import spydrnet as sdn
import spydrnet.global_state.global_callback as gc
from spydrnet.global_state.global_netlist import current_netlist

class TestBundle(unittest.TestCase):
    def setUp(self) -> None:
        self.arg1 = 1
        self.arg2 = 2
        self.kwarg = 3
        self.callcount = 0
        self.callcount2 = 0

    def mycall(self, a1, a2, ka = 0):
        self.callcount += 1
        assert(a1 is self.arg1)
        assert(a2 is self.arg2)
        assert(ka is self.kwarg)
    def mycall2(self, a1, a2, ka = 0):
        self.callcount2 += 1
        assert(a1 is self.arg1)
        assert(a2 is self.arg2)
        assert(ka is self.kwarg)

  
    def call_for_each(self, container, register, call, deregister): 
        '''
        general strategy:
        check to make sure the _container is empty
        register a function
        check to make sure the function is registered
        call the function with some parameters
        check to make sure the function was called with the parameters
        deregister the function
        check to make sure the function was deregistered
        call the function
        check to make sure the funciton was not called
        '''
        assert((current_netlist not in container) or self.mycall not in container[current_netlist])
        assert((current_netlist not in container) or self.mycall2 not in container[current_netlist])
        register(self.mycall)
        assert((current_netlist in container) and self.mycall in container[current_netlist])
        register(self.mycall2)
        assert((current_netlist in container) and self.mycall2 in container[current_netlist])
        call(self.arg1, self.arg2, ka = self.kwarg)
        assert(self.callcount == 1 and self.callcount2 == 1)
        deregister(self.mycall)
        assert((current_netlist not in container) or self.mycall not in container[current_netlist])
        call(self.arg1, self.arg2, ka = self.kwarg)
        assert((self.callcount2 == 2 and self.callcount == 1))
        deregister(self.mycall2)
        assert((current_netlist not in container) or self.mycall2 not in container[current_netlist])
        call(self.arg1, self.arg2, ka = self.kwarg)
        assert(self.callcount == 1 and self.callcount2 == 2)

    def test_cable_add_wire(self):
        self.call_for_each(gc._container_cable_add_wire, gc.register_cable_add_wire, gc._call_cable_add_wire, gc.deregister_cable_add_wire)

    def test_cable_remove_wire(self):
        self.call_for_each(gc._container_cable_remove_wire, gc.register_cable_remove_wire, gc._call_cable_remove_wire, gc.deregister_cable_remove_wire)

    def test_definition_add_port(self):
        self.call_for_each(gc._container_definition_add_port, gc.register_definition_add_port, gc._call_definition_add_port, gc.deregister_definition_add_port)

    def test_definition_remove_port(self):
        self.call_for_each(gc._container_definition_remove_port, gc.register_definition_remove_port, gc._call_definition_remove_port, gc.deregister_definition_remove_port)

    def test_definition_add_child(self):
        self.call_for_each(gc._container_definition_add_child, gc.register_definition_add_child, gc._call_definition_add_child, gc.deregister_definition_add_child)

    def test_definition_remove_child(self):
        self.call_for_each(gc._container_definition_remove_child, gc.register_definition_remove_child, gc._call_definition_remove_child, gc.deregister_definition_remove_child)

    def test_definition_add_cable(self):
        self.call_for_each(gc._container_definition_add_cable, gc.register_definition_add_cable, gc._call_definition_add_cable, gc.deregister_definition_add_cable)

    def test_definition_remove_cable(self):
        self.call_for_each(gc._container_definition_remove_cable, gc.register_definition_remove_cable, gc._call_definition_remove_cable, gc.deregister_definition_remove_cable)

    def test_instance_reference(self):
        self.call_for_each(gc._container_instance_reference, gc.register_instance_reference, gc._call_instance_reference, gc.deregister_instance_reference)

    def test_library_add_definition(self):
        self.call_for_each(gc._container_library_add_definition, gc.register_library_add_definition, gc._call_library_add_definition, gc.deregister_library_add_definition)

    def test_library_remove_definition(self):
        self.call_for_each(gc._container_library_remove_definition, gc.register_library_remove_definition, gc._call_library_remove_definition, gc.deregister_library_remove_definition)

    def test_netlist_top_instance(self):
        self.call_for_each(gc._container_netlist_top_instance, gc.register_netlist_top_instance, gc._call_netlist_top_instance, gc.deregister_netlist_top_instance)

    def test_netlist_add_library(self):
        self.call_for_each(gc._container_netlist_add_library, gc.register_netlist_add_library, gc._call_netlist_add_library, gc.deregister_netlist_add_library)

    def test_netlist_remove_library(self):
        self.call_for_each(gc._container_netlist_remove_library, gc.register_netlist_remove_library, gc._call_netlist_remove_library, gc.deregister_netlist_remove_library)

    def test_port_add_pin(self):
        self.call_for_each(gc._container_port_add_pin, gc.register_port_add_pin, gc._call_port_add_pin, gc.deregister_port_add_pin)

    def test_port_remove_pin(self):
        self.call_for_each(gc._container_port_remove_pin, gc.register_port_remove_pin, gc._call_port_remove_pin, gc.deregister_port_remove_pin)

    def test_wire_connect_pin(self):
        self.call_for_each(gc._container_wire_connect_pin, gc.register_wire_connect_pin, gc._call_wire_connect_pin, gc.deregister_wire_connect_pin)

    def test_wire_disconnect_pin(self):
        self.call_for_each(gc._container_wire_disconnect_pin, gc.register_wire_disconnect_pin, gc._call_wire_disconnect_pin, gc.deregister_wire_disconnect_pin)

    def test_dictionary_delete(self):
        self.call_for_each(gc._container_dictionary_delete, gc.register_dictionary_delete, gc._call_dictionary_delete, gc.deregister_dictionary_delete)

    def test_dictionary_pop(self):
        self.call_for_each(gc._container_dictionary_pop, gc.register_dictionary_pop, gc._call_dictionary_pop, gc.deregister_dictionary_pop)
    
    def test_dictionary_add(self):
        self.call_for_each(gc._container_dictionary_set, gc.register_dictionary_set, gc._call_dictionary_set, gc.deregister_dictionary_set)
        


import unittest
import spydrnet as sdn


class TestOuterPinsView(unittest.TestCase):
    def setUp(self) -> None:
        definition = sdn.Definition()
        port = definition.create_port()
        self.inner_pins = port.create_pins(10)
        self.instance = sdn.Instance()
        self.instance.reference = definition
        self.outer_pins_view = self.instance.pins

    def test_contains(self):
        self.assertTrue(all(x in self.outer_pins_view for x in self.inner_pins))
        self.assertTrue(all(sdn.OuterPin(self.instance, x) in self.outer_pins_view for x in self.inner_pins))

    def test_equal(self):
        self.assertEqual(self.outer_pins_view, dict(map(lambda x: (x, sdn.OuterPin(self.instance, x)),
                                                        self.inner_pins)))

    def test_getitem(self):
        self.assertTrue(all(self.outer_pins_view[x] == sdn.OuterPin(self.instance, x) for x in self.inner_pins))
        self.assertTrue(all(self.outer_pins_view[x] is self.outer_pins_view[sdn.OuterPin(self.instance, x)] for x in
                            self.inner_pins))

    def test_iter(self):
        self.assertTrue(all(isinstance(x, sdn.OuterPin) for x in self.outer_pins_view))

    def test_len(self):
        self.assertEqual(len(self.outer_pins_view), 10)

    def test_get(self):
        self.assertEqual(self.outer_pins_view.get(self.inner_pins[0]), sdn.OuterPin(self.instance, self.inner_pins[0]))
        self.assertEqual(self.outer_pins_view.get(sdn.OuterPin(self.instance, self.inner_pins[0])),
                         sdn.OuterPin(self.instance, self.inner_pins[0]))



import unittest

from spydrnet.ir.element import Element
from spydrnet.ir.pin import Pin


class TestPin(unittest.TestCase):
    def setUp(self):
        self.pin = Pin()

    def test_constructor(self):
        self.assertFalse(isinstance(self.pin, Element))
        self.assertTrue(self.pin)
        pin2 = Pin()
        self.assertNotEqual(self.pin, pin2)

    def test_wire(self):
        self.assertIsNone(self.pin.wire)

    @unittest.expectedFailure
    def test_wire_set(self):
        self.pin.wire = None


import unittest

import spydrnet as sdn
from spydrnet.ir.pin import Pin


class TestInnerPin(unittest.TestCase):
    def setUp(self) -> None:
        self.pin = sdn.InnerPin()

    def test_constructor(self):
        self.assertIsInstance(self.pin, Pin)
        self.assertTrue(self.pin)
        pin2 = sdn.InnerPin()
        self.assertNotEqual(self.pin, pin2)

    def test_port(self):
        self.assertIsNone(self.pin.port)

    @unittest.expectedFailure
    def test_port_set(self):
        self.pin.port = None


import unittest
import os
import io
import zipfile

from spydrnet.parsers.edif.tokenizer import EdifTokenizer
from spydrnet import base_dir

class TestEdifTokenizer(unittest.TestCase):

    def test_no_constructor_of_zero_argument(self):
        self.assertRaises(TypeError, EdifTokenizer)

    def test_stream(self):
        dir_of_edif_netlists = os.path.join(base_dir, "support_files", "EDIF_netlists")
        test_file = os.path.join(dir_of_edif_netlists, "n_bit_counter.edf.zip")
        zip = zipfile.ZipFile(test_file)
        file_name = os.path.basename(test_file)
        file_name = file_name[:file_name.rindex(".")]
        stream = zip.open(file_name)
        stream = io.TextIOWrapper(stream)
        tokenizer = EdifTokenizer.from_stream(stream)
        next_token = tokenizer.next()
        self.assertEqual("(", next_token)

    def test_open_zip_file(self):
        dir_of_edif_netlists = os.path.join(base_dir, "support_files", "EDIF_netlists")
        test_file = os.path.join(dir_of_edif_netlists, "n_bit_counter.edf.zip")
        tokenizer = EdifTokenizer.from_filename(test_file)
        next_token = tokenizer.next()
        self.assertEqual("(", next_token)

    def test_open_file(self):
        dir_of_edif_netlists = os.path.join(base_dir, "support_files", "EDIF_netlists")
        test_file = os.path.join(dir_of_edif_netlists, "n_bit_counter.edf.zip")
        file_name = os.path.basename(test_file)
        file_name = file_name[:file_name.rindex(".")]
        extract_path = os.path.join(dir_of_edif_netlists, file_name)
        if os.path.exists(extract_path):
            os.remove(extract_path)
        zip = zipfile.ZipFile(test_file)
        zip.extract(file_name, dir_of_edif_netlists)
        tokenizer = EdifTokenizer.from_filename(extract_path)
        next_token = tokenizer.next()
        self.assertEqual("(", next_token)
        tokenizer.close()
        os.remove(extract_path)

    def test_empty_string(self):
        tokenizer = EdifTokenizer.from_string("")
        self.assertRaises(StopIteration, tokenizer.next)

    def test_has_next_false(self):
        tokenizer = EdifTokenizer.from_string("")
        self.assertFalse(tokenizer.has_next())

    def test_has_next_true(self):
        tokenizer = EdifTokenizer.from_string("(")
        self.assertTrue(tokenizer.has_next())

    def test_paren_with_whitespace(self):
        tokenizer = EdifTokenizer.from_string(" \t\n\r( \r\n\t )\n\r \t")
        self.assertEqual("(", tokenizer.next())
        self.assertEqual(2, tokenizer.line_number)
        self.assertEqual(")", tokenizer.next())
        self.assertEqual(3, tokenizer.line_number)
        self.assertRaises(StopIteration, tokenizer.next)

    def test_string_token(self):
        test_string = "\"This is a test string in EDIF\""
        tokenizer = EdifTokenizer.from_string(test_string)
        next_token = tokenizer.next()
        self.assertEqual(test_string, next_token)
        self.assertRaises(StopIteration, tokenizer.next)

    def test_multiline_string(self):
        test_string = "\"This is a test string in EDIF\n with a new line in it.\""
        tokenizer = EdifTokenizer.from_string(test_string)
        next_token = tokenizer.next()
        self.assertEqual(test_string.replace("\n", ""), next_token)
        self.assertRaises(StopIteration, tokenizer.next)
    
    def test_peek_and_token_equals(self):
        test_id = "VALID_EDIF_ID"
        tokenizer = EdifTokenizer.from_string(test_id)
        self.assertTrue(tokenizer.has_next())
        self.assertFalse(tokenizer.peek_equals("DIFFERENT_TOKEN"))
        self.assertTrue(tokenizer.peek_equals("valid_edif_id"))
        self.assertTrue(tokenizer.peek_equals("vAlId_eDIf_ID"))
        self.assertTrue(tokenizer.peek_equals(test_id))
        tokenizer.next()
        self.assertFalse(tokenizer.token_equals("DIFFERENT_TOKEN"))
        self.assertTrue(tokenizer.token_equals("valid_edif_id"))
        self.assertTrue(tokenizer.token_equals("vAlId_eDIf_ID"))
        self.assertTrue(tokenizer.token_equals(test_id))


class TestTokenTypes(unittest.TestCase):

    def test_edif_identifier(self):
        valid_ids = "&test a a9 a_ Alpha_Numeric_0123456789 &"
        tokenizer = EdifTokenizer.from_string(valid_ids)
        while tokenizer.has_next():
            tokenizer.next()
            self.assertTrue(tokenizer.is_valid_identifier(), tokenizer.token)

        invalid_ids = "_ _idenitier_ 9alkjf too_long" + "0"*(256 - 7)
        tokenizer = EdifTokenizer.from_string(invalid_ids)
        while tokenizer.has_next():
            tokenizer.next()
            self.assertFalse(tokenizer.is_valid_identifier(), tokenizer.token)

        #TODO Special characters are weird. We should check them out.


if __name__ == '__main__':
    unittest.main()


class ListView:
    __slots__ = ['_list']

    def __init__(self, list_object):
        self._list = list_object

    def __add__(self, other):
        return self._list.__add__(other)

    def __getitem__(self, other):
        return self._list.__getitem__(other)

    def __contains__(self, item):
        return self._list.__contains__(item)

    def __eq__(self, other):
        return self._list.__eq__(other)

    def __hash__(self):
        return self._list.__hash__()

    def __ge__(self, other):
        if isinstance(other, ListView):
            return self._list.__ge__(other._list)
        return self._list.__ge__(other)

    def __gt__(self, other):
        if isinstance(other, ListView):
            return self._list.__gt__(other._list)
        return self._list.__gt__(other)

    def __iter__(self):
        return self._list.__iter__()

    def __le__(self, other):
        if isinstance(other, ListView):
            return self._list.__le__(other._list)
        return self._list.__le__(other)

    def __len__(self):
        return self._list.__len__()

    def __lt__(self, other):
        if isinstance(other, ListView):
            return self._list.__lt__(other._list)
        return self._list.__lt__(other)

    def __ne__(self, other):
        return self._list.__ne__(other)

    def __mul__(self, other):
        return self._list.__mul__(other)

    def __rmul__(self, n):
        return self._list.__rmul__(n)

    def __reversed__(self):
        return self._list.__reversed__()

    def __repr__(self):
        return self._list.__repr__()

    def __str__(self):
        return self._list.__str__()

    def __radd__(self, other):
        return other + self._list

    def __iadd__(self, other):
        raise TypeError("unsupported operator for type SetView")

    def __imul__(self, other):
        raise TypeError("unsupported operator for type SetView")

    def copy(self):
        return self._list.copy()

    def count(self, object):
        return self._list.count(object)

    def index(self, *args, **kwargs):
        return self._list.index(*args, **kwargs)


import unittest
import spydrnet as sdn
import glob
import os


class TestExampleNetlistFunctionality(unittest.TestCase):
    def test_example_netlist_names(self):
        filenames = glob.glob(os.path.join(sdn.base_dir, 'support_files', 'EDIF_netlists', '*'))
        self.assertEqual(len(filenames), len(sdn.example_netlist_names))
        filenames.sort()
        for filename, example_name in zip(filenames, sdn.example_netlist_names):
            basename = os.path.basename(filename)
            example_name_golden = basename[:basename.index('.')]
            self.assertEqual(example_name, example_name_golden)

    def test_load_example_netlist_by_name(self):
        netlist = sdn.load_example_netlist_by_name(sdn.example_netlist_names[0])
        self.assertIsNotNone(netlist)



class SetView:
    __slots__ = ['_set']

    def __init__(self, set_object):
        self._set = set_object

    def __and__(self, *args, **kwargs):
        return self._set.__and__(*args, *kwargs)

    def __rand__(self, *args, **kwargs):
        return self._set.__rand__(*args, **kwargs)

    def __eq__(self, other):
        if isinstance(other, SetView):
            return self._set.__eq__(other._set)
        return self._set.__eq__(other)

    def __ge__(self, other):
        if isinstance(other, SetView):
            return self._set.__ge__(other._set)
        return self._set.__ge__(other)

    def __gt__(self, other):
        if isinstance(other, SetView):
            return self._set.__gt__(other._set)
        return self._set.__gt__(other)

    def __iter__(self, *args, **kwargs):
        return self._set.__iter__(*args, **kwargs)

    def __le__(self, other):
        if isinstance(other, SetView):
            return self._set.__le__(other._set)
        return self._set.__le__(other)

    def __len__(self, *args, **kwargs):
        return self._set.__len__(*args, **kwargs)

    def __lt__(self, other):
        if isinstance(other, SetView):
            return self._set.__lt__(other._set)
        return self._set.__lt__(other)

    def __ne__(self, other):
        if isinstance(other, SetView):
            return self._set.__ne__(other._set)
        return self._set.__ne__(other)

    def __or__(self, *args, **kwargs):
        return self._set.__or__(*args, **kwargs)

    def __ror__(self, *args, **kwargs):
        return self._set.__ror__(*args, **kwargs)

    def __sub__(self, *args, **kwargs):
        return self._set.__sub__(*args, **kwargs)

    def __rsub__(self, *args, **kwargs):
        return self._set.__rsub__(*args, **kwargs)

    def __xor__(self, *args, **kwargs):
        return self._set.__xor__(*args, **kwargs)

    def __rxor__(self, *args, **kwargs):
        return self._set.__rxor__(*args, **kwargs)

    def __repr__(self, *args, **kwargs):
        return self._set.__repr__(*args, **kwargs)

    def __str__(self, *args, **kwargs):
        return self._set.__str__(*args, **kwargs)

    def copy(self, *args, **kwargs):
        return self._set.copy(*args, **kwargs)

    def difference(self, *args, **kwargs):
        return self._set.difference(*args, **kwargs)

    def intersection(self, *args, **kwargs):
        return self._set.intersection(*args, **kwargs)

    def isdisjoint(self, *args, **kwargs):
        return self._set.isdisjoint(*args, **kwargs)

    def issubset(self, *args, **kwargs):
        return self._set.issubset(*args, **kwargs)

    def issuperset(self, *args, **kwargs):
        return self._set.issuperset(*args, **kwargs)

    def symmetric_difference(self, *args, **kwargs):
        return self._set.symmetric_difference(*args, **kwargs)

    def union(self, *args, **kwargs):
        return self._set.union(*args, **kwargs)

    def __iand__(self, *args, **kwargs):
        raise TypeError("unsupported operator for type SetView")

    def __ior__(self, *args, **kwargs):
        raise TypeError("unsupported operator for type SetView")

    def __ixor__(self, *args, **kwargs):
        raise TypeError("unsupported operator for type SetView")

    def __isub__(self, *args, **kwargs):
        raise TypeError("unsupported operator for type SetView")



import os

# Version information
name = 'spydrnet'
version = "UNSPECIFIED"
date = "UNSPECIFIED"

# Declare current release as a development release.
# Change to False before tagging a release; then change back.
dev = False

description = "Python package for analyzing and transforming netlists"
license = 'BSD'
authors = {'Keller': ('Andrew Keller', 'andrewmkeller@byu.edu'),
           'Skouson': ('Dallin Skouson', 'dallinskouson@byu.edu'),
           'Wirthlin': ('Michael Wirthlin', 'wirthlin@byu.edu')}
maintainer = "SpyDrNet Developers"
maintainer_email = "spydrnet-discuss@googlegroups.com"
url = 'https://byuccl.github.io/spydrnet'
project_urls = {
    "Bug Tracker": "https://github.com/byuccl/spydrnet/issues",
    "Documentation": "https://byuccl.github.io/spydrnet/docs/stable",
    "Source Code": "https://github.com/byuccl/spydrnet",
}
platforms = ['Linux', 'Mac OSX', 'Windows', 'Unix']
keywords = ['Netlist', 'Analysis', 'Transformation',
            'netlist', 'FPGA', 'primitives', 'module', 
            'port', 'EDIF', 'Digital', 'Hardware',]
classifiers = [
            "Development Status :: 5 - Production/Stable",
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: BSD License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3 :: Only",
            "Topic :: Education",
            "Topic :: Scientific/Engineering",
            'Topic :: Scientific/Engineering :: Information Analysis',
            "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
            "Topic :: Software Development :: Libraries :: Python Modules"
        ]


directory = os.path.dirname(os.path.abspath(__file__))
version_filename = os.path.join(directory, "VERSION")

def load_versionfile():
    global version
    global date
    if os.path.isfile(version_filename):
        with open(version_filename) as fi:
            version = fi.readline().strip()[1:]
            date = fi.readline().strip()
            
load_versionfile()

def update_versionfile():
    import datetime
    import time
    import subprocess
    
    global version
    global date
    
    date_info = datetime.datetime.utcfromtimestamp(int(os.environ.get('SOURCE_DATE_EPOCH', time.time())))
    date = time.asctime(date_info.timetuple())

    #check for git on path
    git_exists = False
    for path in os.environ['PATH'].split(os.pathsep):
        exe_file = os.path.join(path, 'git')
        if os.path.isfile(exe_file) and os.access(exe_file, os.X_OK):
            git_exists = True
        else:
            exe_file = os.path.join(path, 'git.exe')
            if os.path.isfile(exe_file) and os.access(exe_file, os.X_OK):
                git_exists = True
            
    if git_exists:
        git_subprocess = subprocess.Popen(["git", "describe"], shell=True, cwd=directory, stdout=subprocess.PIPE)
        git_describe_output = git_subprocess.stdout.read().decode()

        git_version = git_describe_output.strip()
        if git_version.startswith('v'):
            version_file = os.path.join(directory, "VERSION")
            with open(version_file, 'w') as fh:
                fh.write(git_version + '\n')
                fh.write(date + '\n')
            version = git_version[1:]
            
    return version
    
if __name__ == '__main__':
    update_versionfile()


import unittest
import spydrnet as sdn

class TestComposers(unittest.TestCase):
    def test_compose(self):
        netlist = sdn.Netlist()
        self.assertRaises(RuntimeError, sdn.compose, "fakefile.fakeext", netlist)



#Parse primative library files from various vendors.
#Copyright 2019 BYU Configurable Computing Lab
#Author Dallin Skouson
#See the license that accompanies SpyDrNet tools for more information
#
#Currently implemented vendors:
#Xilinx vivado tested with and designed for files from Vivado/2019.1/data/verilog/src/xeclib (other versions should hopefully work)
#
#Future capabilities
#accept files from other locations for Xilinx
#support intel/altera devices.
#support for other vendors

import os
import SpyDrNet

VERBOSE = True

class PrimativeParser:
    '''
    Parses primative libraries for use with the tools

    This class is a standalone class as of the first release. 
    Hopefully the xilinx part of this parser can be combined with the structural verilog parser so that they can share code

    Extracts the ports, parameters, and names of primatives
    The primative libraries can be found in the following location
    Xilinx Vivado:
    Vivado/2019.1/data/verilog/src/xeclib
    '''

    def vivado_parse_xeclib(self, xeclib_directory):
        '''
        parse each file in the xeclib_directory
        parameters:
            xeclib_directory - The directory where the primatives can be found
        returns:
            Library that contains all the primatives and is orphaned from any netlist.
        '''
        for v_file in os.listdir(xeclib_directory):
            self._vivado_parse_file(xeclib_directory + "/" + v_file)

    def _vivado_parse_file(self, verilog_file):
        '''
        parse the verilog file
        parameters:
            a verilog file to parse
        returns:
            Definition created from the verilog file
        '''
        if not verilog_file.endswith('.v'):
           print("File " + verilog_file + " in directory does not end with the proper extension") 
        with open(verilog_file) as vf:
            self._vivado_parse(vf)

    def _vivado_parse(self, stream):
        parameters = {}
        inputs = {}
        outputs = {}
        inouts = {}
        name = ''
        for statement in stream:
            statement = statement.rstrip()
            if statement.endswith(';') or statement.endswith(','):
                statement = statement[:-1]
            statement = statement.split()
            if len(statement) == 0:
                continue
            elif statement[0] == 'parameter':
                pass
            elif statement[0] == 'module':
                name = statement[1] #this is lazy but I think it might work with all the xilinx verilog files for now....
            elif statement[0] == 'input':
                self._vivado_get_port_info(statement, inputs)
            elif statement[0] == 'output':
                self._vivado_get_port_info(statement, outputs)
            elif statement[0] == 'inout':
                self._vivado_get_port_info(statement, inouts)
            else:
                pass

        if(VERBOSE):
            print("\nModule " + name)
            print("Inputs:")
            print(inputs)
            print("Outputs:")
            print(outputs)
            print("Inouts:")
            print(inouts)

        definition = self._vivado_create_definition(name, parameters, inputs, outputs, inouts)
        return definition

    def _vivado_get_port_info(self, line, port_container):
        if line[1][0] == '[':
            direction = line[1]
            direction = direction[1:-1]
            from_to = direction.split(":")
            from_to[0] = int(from_to[0])
            from_to[1] = int(from_to[1])
            for i in range(2,len(line)):
                port_container[line[i]] = from_to
        else:
            for i in range(1,len(line)):
                port_container[line[i]] = ''
    
    def _vivado_create_definition(self, name, parameters, inputs, outputs, inouts):
        definition = SpyDrNet.create_definition()
        for i, val in inputs.items():
            definition.add_port(self._create_port(i,val,SpyDrNet.Port.Direction.IN))
        for o, val in outputs.items():
            definition.add_port(self._create_port(o,val,SpyDrNet.Port.Direction.OUT))
        for io, val in inouts.items():
            definition.add_port(self._create_port(io,val,SpyDrNet.Port.Direction.INOUT))
        return definition
        
    def _create_port(self, name, indicies, direction):
        port = SpyDrNet.create_port(name = name)
        port.set_direction(direction)
        if indicies != '':
            port.initialize_pins_in_range(indicies[0],indicies[1])
        return port

if __name__ == "__main__":
    parse = PrimativeParser()
    parse.vivado_parse_xeclib("/home/dallin/Documents/byuccl/xeclib")


from spydrnet.global_state.global_netlist import get_current_netlist

_container_cable_add_wire = dict()
_container_cable_remove_wire = dict()
_container_definition_add_port = dict()
_container_definition_remove_port = dict()
_container_definition_add_child = dict()
_container_definition_remove_child = dict()
_container_definition_add_cable = dict()
_container_definition_remove_cable = dict()
_container_instance_reference = dict()
_container_library_add_definition = dict()
_container_library_remove_definition = dict()
_container_netlist_top_instance = dict()
_container_netlist_add_library = dict()
_container_netlist_remove_library = dict()
_container_port_add_pin = dict()
_container_port_remove_pin = dict()
_container_wire_connect_pin = dict()
_container_wire_disconnect_pin = dict()
#dictionary containers
_container_dictionary_set = dict()
_container_dictionary_delete = dict()
_container_dictionary_pop = dict()

#look into inlining this function perhaps
def _call(dictionary_to_call, *args, **kwargs):
    if get_current_netlist() in dictionary_to_call:
        for func in dictionary_to_call[get_current_netlist()]:
            func(*args, **kwargs)

def _call_cable_add_wire(*args, **kwargs):
    _call(_container_cable_add_wire, *args, **kwargs)

def _call_cable_remove_wire(*args, **kwargs):
    _call(_container_cable_remove_wire, *args, **kwargs)

def _call_definition_add_port(*args, **kwargs):
    _call(_container_definition_add_port, *args, **kwargs)

def _call_definition_remove_port(*args, **kwargs):
    _call(_container_definition_remove_port, *args, **kwargs)

def _call_definition_add_child(*args, **kwargs):
    _call(_container_definition_add_child, *args, **kwargs)

def _call_definition_remove_child(*args, **kwargs):
    _call(_container_definition_remove_child, *args, **kwargs)

def _call_definition_add_cable(*args, **kwargs):
    _call(_container_definition_add_cable, *args, **kwargs)

def _call_definition_remove_cable(*args, **kwargs):
    _call(_container_definition_remove_cable, *args, **kwargs)

def _call_instance_reference(*args, **kwargs):
    _call(_container_instance_reference, *args, **kwargs)

def _call_library_add_definition(*args, **kwargs):
    _call(_container_library_add_definition, *args, **kwargs)

def _call_library_remove_definition(*args, **kwargs):
    _call(_container_library_remove_definition, *args, **kwargs)

def _call_netlist_top_instance(*args, **kwargs):
    _call(_container_netlist_top_instance, *args, **kwargs)

def _call_netlist_add_library(*args, **kwargs):
    _call(_container_netlist_add_library, *args, **kwargs)

def _call_netlist_remove_library(*args, **kwargs):
    _call(_container_netlist_remove_library, *args, **kwargs)

def _call_port_add_pin(*args, **kwargs):
    _call(_container_port_add_pin, *args, **kwargs)

def _call_port_remove_pin(*args, **kwargs):
    _call(_container_port_remove_pin, *args, **kwargs)

def _call_wire_connect_pin(*args, **kwargs):
    _call(_container_wire_connect_pin, *args, **kwargs)

def _call_wire_disconnect_pin(*args, **kwargs):
    _call(_container_wire_disconnect_pin, *args, **kwargs)

#dictionary functions

def _call_dictionary_set(*args, **kwargs):
    _call(_container_dictionary_set, *args, **kwargs)

def _call_dictionary_delete(*args, **kwargs):
    _call(_container_dictionary_delete, *args, **kwargs)

def _call_dictionary_pop(*args, **kwargs):
    _call(_container_dictionary_pop, *args, **kwargs)

'''register functions'''

#look into inlining this function perhaps
def _register(dictionary_to_register, method):
    if get_current_netlist() in dictionary_to_register:
        assert(method not in dictionary_to_register[get_current_netlist()])
        dictionary_to_register[get_current_netlist()].append(method)
    else:
        dictionary_to_register[get_current_netlist()] = []
        dictionary_to_register[get_current_netlist()].append(method)

def register_cable_add_wire(method):
    _register(_container_cable_add_wire, method)

def register_cable_remove_wire(method):
    _register(_container_cable_remove_wire, method)

def register_definition_add_port(method):
    _register(_container_definition_add_port, method)

def register_definition_remove_port(method):
    _register(_container_definition_remove_port, method)

def register_definition_add_child(method):
    _register(_container_definition_add_child, method)

def register_definition_remove_child(method):
    _register(_container_definition_remove_child, method)

def register_definition_add_cable(method):
    _register(_container_definition_add_cable, method)

def register_definition_remove_cable(method):
    _register(_container_definition_remove_cable, method)

def register_instance_reference(method):
    _register(_container_instance_reference, method)

def register_library_add_definition(method):
    _register(_container_library_add_definition, method)

def register_library_remove_definition(method):
    _register(_container_library_remove_definition, method)

def register_netlist_top_instance(method):
    _register(_container_netlist_top_instance, method)

def register_netlist_add_library(method):
    _register(_container_netlist_add_library, method)

def register_netlist_remove_library(method):
    _register(_container_netlist_remove_library, method)

def register_port_add_pin(method):
    _register(_container_port_add_pin, method)

def register_port_remove_pin(method):
    _register(_container_port_remove_pin, method)

def register_wire_connect_pin(method):
    _register(_container_wire_connect_pin, method)

def register_wire_disconnect_pin(method):
    _register(_container_wire_disconnect_pin, method)

def register_dictionary_set(*args, **kwargs):
    _register(_container_dictionary_set, *args, **kwargs)

def register_dictionary_delete(*args, **kwargs):
    _register(_container_dictionary_delete, *args, **kwargs)

def register_dictionary_pop(*args, **kwargs):
    _register(_container_dictionary_pop, *args, **kwargs)

'''deregister functions'''

#look into inlining this function perhaps
def _deregister(dictionary_to_deregister, method):
    assert(get_current_netlist() in dictionary_to_deregister)
    assert(method in dictionary_to_deregister[get_current_netlist()])
    dictionary_to_deregister[get_current_netlist()].remove(method)

def deregister_cable_add_wire(method):
    _deregister(_container_cable_add_wire, method)

def deregister_cable_remove_wire(method):
    _deregister(_container_cable_remove_wire, method)

def deregister_definition_add_port(method):
    _deregister(_container_definition_add_port, method)

def deregister_definition_remove_port(method):
    _deregister(_container_definition_remove_port, method)

def deregister_definition_add_child(method):
    _deregister(_container_definition_add_child, method)

def deregister_definition_remove_child(method):
    _deregister(_container_definition_remove_child, method)

def deregister_definition_add_cable(method):
    _deregister(_container_definition_add_cable, method)

def deregister_definition_remove_cable(method):
    _deregister(_container_definition_remove_cable, method)

def deregister_instance_reference(method):
    _deregister(_container_instance_reference, method)

def deregister_library_add_definition(method):
    _deregister(_container_library_add_definition, method)

def deregister_library_remove_definition(method):
    _deregister(_container_library_remove_definition, method)

def deregister_netlist_top_instance(method):
    _deregister(_container_netlist_top_instance, method)

def deregister_netlist_add_library(method):
    _deregister(_container_netlist_add_library, method)

def deregister_netlist_remove_library(method):
    _deregister(_container_netlist_remove_library, method)

def deregister_port_add_pin(method):
    _deregister(_container_port_add_pin, method)

def deregister_port_remove_pin(method):
    _deregister(_container_port_remove_pin, method)

def deregister_wire_connect_pin(method):
    _deregister(_container_wire_connect_pin, method)

def deregister_wire_disconnect_pin(method):
    _deregister(_container_wire_disconnect_pin, method)

def deregister_dictionary_set(*args, **kwargs):
    _deregister(_container_dictionary_set, *args, **kwargs)

def deregister_dictionary_delete(*args, **kwargs):
    _deregister(_container_dictionary_delete, *args, **kwargs)

def deregister_dictionary_pop(*args, **kwargs):
    _deregister(_container_dictionary_pop, *args, **kwargs)



from spydrnet.ir.netlist import Netlist
from spydrnet.ir.library import Library
from spydrnet.ir.definition import Definition
from spydrnet.ir.port import Port
from spydrnet.ir.cable import Cable
from spydrnet.ir.wire import Wire
from spydrnet.ir.instance import Instance
from spydrnet.ir.innerpin import InnerPin
from spydrnet.ir.outerpin import OuterPin
from spydrnet.ir.bundle import Bundle
from spydrnet.ir.pin import Pin
from spydrnet.ir.element import Element


from spydrnet.global_state import global_callback


class CallbackListener:
    '''
    extend this class to create a listener that can be used to make a plugin to spydrnet.
    In this class are all of the functions that are used as callbacks. There are also
    register functions whos function is to register the callback functions that are present.
    
    callbacks are netlist dependent. If a callback is registered to one netlist it is not automatically registered to all netlists.
    '''

    def __init__(self):
        self.register_all_listeners()
        
    def register_all_listeners(self):
        '''can be called when needed'''
        '''override this function to call all the register functions that you want to listen for'''
        pass

    def cable_add_wire(self, cable, wire):
        raise NotImplementedError

    def cable_remove_wire(self, cable, wire):
        raise NotImplementedError

    def definition_add_port(self, definition, port):
        raise NotImplementedError

    def definition_remove_port(self, definition, port):
        raise NotImplementedError

    def definition_add_child(self, definition, child):
        raise NotImplementedError

    def definition_remove_child(self, definition, child):
        raise NotImplementedError

    def definition_add_cable(self, definition, cable):
        raise NotImplementedError

    def definition_remove_cable(self, definition, cable):
        raise NotImplementedError

    def instance_reference(self, instance, reference):
        raise NotImplementedError

    def library_add_definition(self, library, definition):
        raise NotImplementedError

    def library_remove_definition(self, library, definition):
        raise NotImplementedError

    def netlist_top_instance(self, netlist, instance):
        raise NotImplementedError

    def netlist_add_library(self, netlist, library):
        raise NotImplementedError

    def netlist_remove_library(self, netlist, library):
        raise NotImplementedError

    def port_add_pin(self, port, pin):
        raise NotImplementedError

    def port_remove_pin(self, port, pin):
        raise NotImplementedError

    def wire_connect_pin(self, wire, pin):
        raise NotImplementedError

    def wire_disconnect_pin(self, wire, pin):
        raise NotImplementedError

    #dictionary watchers
    
    def dictionary_set(self, element, key, value):
        raise NotImplementedError

    def dictionary_delete(self, element, key):
        raise NotImplementedError

    def dictionary_pop(self, element, item):
        raise NotImplementedError


    '''
    call the register_ prefixed methods in your register_all_listners_function.
    '''
                
    def register_cable_add_wire(self):
        global_callback.register_cable_add_wire(self.cable_add_wire)

    def register_cable_remove_wire(self):
        global_callback.register_cable_remove_wire(self.cable_remove_wire)

    def register_definition_add_port(self):
        global_callback.register_definition_add_port(self.definition_add_port)

    def register_definition_remove_port(self):
        global_callback.register_definition_remove_port(self.definition_remove_port)

    def register_definition_add_child(self):
        global_callback.register_definition_add_child(self.definition_add_child)

    def register_definition_remove_child(self):
        global_callback.register_definition_remove_child(self.definition_remove_child)

    def register_definition_add_cable(self):
        global_callback.register_definition_add_cable(self.definition_add_cable)

    def register_definition_remove_cable(self):
        global_callback.register_definition_remove_cable(self.definition_remove_cable)

    def register_instance_reference(self):
        global_callback.register_instance_reference(self.instance_reference)

    def register_library_add_definition(self):
        global_callback.register_library_add_definition(self.library_add_definition)

    def register_library_remove_definition(self):
        global_callback.register_library_remove_definition(self.library_remove_definition)

    def register_netlist_top_instance(self):
        global_callback.register_netlist_top_instance(self.netlist_top_instance)

    def register_netlist_add_library(self):
        global_callback.register_netlist_add_library(self.netlist_add_library)

    def register_netlist_remove_library(self):
        global_callback.register_netlist_remove_library(self.netlist_remove_library)

    def register_port_add_pin(self):
        global_callback.register_port_add_pin(self.port_add_pin)

    def register_port_remove_pin(self):
        global_callback.register_port_remove_pin(self.port_remove_pin)

    def register_wire_connect_pin(self):
        global_callback.register_wire_connect_pin(self.wire_connect_pin)

    def register_wire_disconnect_pin(self):
        global_callback.register_wire_disconnect_pin(self.wire_disconnect_pin)

    def register_dictionary_set(self):
        global_callback.register_dictionary_set(self.dictionary_set)

    def register_dictionary_delete(self):
        global_callback.register_dictionary_delete(self.dictionary_delete)

    def register_dictionary_pop(self):
        global_callback.register_dictionary_pop(self.dictionary_pop)


class Pin:
    '''pin connects to a single wire. This class is extended by InnerPin and OuterPin'''
    __slots__ = ['_wire']

    def __init__(self):
        self._wire = None

    @property
    def wire(self):
        '''get the wire the pin is connected to. This value cannot be modified directly by the end user.'''
        return self._wire



import os
import zipfile
import tempfile


def parse(filename):
    basename_less_final_extension = os.path.splitext(os.path.basename(filename))[0]
    extension = get_lowercase_extension(filename)
    if extension == ".zip":
        assert zipfile.is_zipfile(filename), \
            "Input filename {} with extension .zip is not a zip file.".format(basename_less_final_extension)
        with tempfile.TemporaryDirectory() as tempdirname:
            with zipfile.ZipFile(filename) as zip:
                files = zip.namelist()
                assert len(files) == 1 and files[0] == basename_less_final_extension, \
                    "Only single file archives allowed with a file whose name matches the name of the archive"
                zip.extract(basename_less_final_extension, tempdirname)
                filename = os.path.join(tempdirname, basename_less_final_extension)
                return _parse(filename)
    return _parse(filename)


def _parse(filename):
    extension = get_lowercase_extension(filename)
    if extension in [".edf", ".edif"]:
        from spydrnet.parsers.edif.parser import EdifParser
        parser = EdifParser.from_filename(filename)
    else:
        raise RuntimeError("Extension {} not recognized.".format(extension))
    parser.parse()
    return parser.netlist

def get_lowercase_extension(filename):
    extension = os.path.splitext(filename)[1]
    extension_lower = extension.lower()
    return extension_lower


import unittest

import spydrnet as sdn
from spydrnet.ir.element import Element


class TestWire(unittest.TestCase):
    def setUp(self):
        self.definition_top = sdn.Definition()
        self.port_top = self.definition_top.create_port()
        self.inner_pin = self.port_top.create_pin()
        self.cable = self.definition_top.create_cable()
        self.wire = self.cable.create_wire()
        self.definition_leaf = sdn.Definition()
        self.port = self.definition_leaf.create_port()
        self.pin1 = self.port.create_pin()
        self.pin2 = self.port.create_pin()
        self.instance = self.definition_top.create_child()
        self.instance.reference = self.definition_leaf

    def test_constructor(self):
        self.assertFalse(isinstance(self.wire, Element), "Wire should not extend element")
        wire2 = sdn.Wire()
        self.assertNotEqual(self.wire, wire2, "Unique items are considered equal")

    def test_pins_assignement(self):
        self.wire.connect_pin(self.instance.pins[self.pin1])
        self.wire.connect_pin(self.instance.pins[self.pin2])
        self.assertEqual(self.wire.pins, [self.instance.pins[self.pin1], self.instance.pins[self.pin2]])
        self.wire.pins = [self.instance.pins[self.pin2], self.instance.pins[self.pin1]]
        self.assertEqual(self.wire.pins, [self.instance.pins[self.pin2], self.instance.pins[self.pin1]])

    def test_connect_and_disconnect_inner_port(self):
        self.wire.connect_pin(self.inner_pin)
        self.assertTrue(self.inner_pin in self.wire.pins)
        self.assertEqual(self.inner_pin.wire, self.wire)
        self.assertEqual(len(self.wire.pins), 1)

        self.wire.disconnect_pin(self.inner_pin)
        self.assertFalse(self.inner_pin in self.wire.pins)
        self.assertIsNone(self.inner_pin.wire)
        self.assertEqual(len(self.wire.pins), 0)

    def test_connect_and_disconnect_outer_pin_by_reference(self):
        self.wire.connect_pin(self.instance.pins[self.pin1])
        self.assertEqual(len(self.wire.pins), 1)
        self.assertTrue(all(x is self.instance.pins[x] for x in self.wire.pins))
        self.assertTrue(all(x.wire is self.wire for x in self.wire.pins))
        self.assertTrue(all(x.instance is self.instance for x in self.wire.pins))
        self.assertEqual(self.instance.pins[self.pin1].inner_pin, self.pin1)

        self.wire.disconnect_pin(self.instance.pins[self.pin1])
        self.assertEqual(len(self.wire.pins), 0)
        self.assertFalse(self.instance.pins[self.pin1] in self.wire.pins)
        self.assertIsNone(self.instance.pins[self.pin1].wire)
        self.assertTrue(self.pin1 in self.instance.pins)

    def test_connect_and_disconnect_outer_pin_by_object(self):
        self.wire.connect_pin(sdn.OuterPin.from_instance_and_inner_pin(self.instance, self.pin2), position=0)
        self.assertEqual(len(self.wire.pins), 1)
        self.assertTrue(all(x is self.instance.pins[x] for x in self.wire.pins))
        self.assertTrue(all(x.wire is self.wire for x in self.wire.pins))
        self.assertTrue(all(x.instance is self.instance for x in self.wire.pins))
        self.assertEqual(self.instance.pins[self.pin2].inner_pin, self.pin2)

        self.wire.disconnect_pin(sdn.OuterPin(self.instance, self.pin2))
        self.assertEqual(len(self.wire.pins), 0)
        self.assertFalse(self.instance.pins[self.pin2] in self.wire.pins)
        self.assertIsNone(self.instance.pins[self.pin1].wire)
        self.assertTrue(self.pin1 in self.instance.pins)

    def test_disconnect_pin_from(self):
        self.wire.connect_pin(self.inner_pin)
        self.wire.connect_pin(self.instance.pins[self.pin1])
        self.wire.connect_pin(self.instance.pins[self.pin2])
        self.wire.disconnect_pins_from(iter((self.inner_pin, self.instance.pins[self.pin1])))
        self.wire.disconnect_pins_from({self.instance.pins[self.pin2]})
        self.assertEqual(len(self.wire.pins), 0)
        self.assertTrue(self.pin1 in self.instance.pins and isinstance(self.instance.pins[self.pin1], sdn.OuterPin) and
                        self.instance.pins[self.pin1].inner_pin == self.pin1)
        self.assertIsNone(self.inner_pin.wire)
        self.assertIsNone(self.instance.pins[self.pin1].wire)
        self.assertIsNone(self.instance.pins[self.pin2].wire)
        self.assertTrue(self.pin1 in self.instance.pins and isinstance(self.instance.pins[self.pin2], sdn.OuterPin) and
                        self.instance.pins[self.pin2].inner_pin == self.pin2)

    @unittest.expectedFailure
    def test_disconnect_inner_pin_from_outside_wire(self):
        inner_pin = sdn.InnerPin()
        self.wire.disconnect_pins_from([inner_pin])

    @unittest.expectedFailure
    def test_disconnect_outer_pin_from_outside_wire(self):
        outer_pin = sdn.OuterPin()
        self.wire.disconnect_pins_from([outer_pin])


import unittest

import spydrnet as sdn
from spydrnet.ir.pin import Pin


class TestOuterPin(unittest.TestCase):
    def setUp(self) -> None:
        self.pin = sdn.OuterPin()

    def test_constructor(self):
        self.assertIsInstance(self.pin, Pin)
        self.assertTrue(self.pin)

    def test_equal(self):
        outer_pin = sdn.OuterPin()
        self.assertEqual(outer_pin, self.pin)
        inner_pin = sdn.InnerPin()
        instance = sdn.Instance()
        outer_pin1 = sdn.OuterPin.from_instance_and_inner_pin(instance, inner_pin)
        outer_pin2 = sdn.OuterPin.from_instance_and_inner_pin(instance, inner_pin)
        self.assertEqual(outer_pin1, outer_pin2)
        self.assertNotEqual(self.pin, outer_pin1)
        self.assertNotEqual(self.pin, None)

    def test_hash(self):
        outer_pin = sdn.OuterPin()
        self.assertEqual(hash(outer_pin), hash(self.pin))
        inner_pin = sdn.InnerPin()
        instance = sdn.Instance()
        outer_pin1 = sdn.OuterPin.from_instance_and_inner_pin(instance, inner_pin)
        outer_pin2 = sdn.OuterPin.from_instance_and_inner_pin(instance, inner_pin)
        self.assertEqual(hash(outer_pin1), hash(outer_pin2))


from spydrnet.ir.bundle import Bundle
from spydrnet.ir.innerpin import InnerPin
from spydrnet.ir.outerpin import OuterPin
from spydrnet.ir.views.listview import ListView
from spydrnet.global_state import global_callback

from enum import Enum


class Port(Bundle):
    '''
    Located on the inside of a definition. Ports contain information about the quantity and directon of pins that go into and out of the defined struture when instanced.
    '''
    __slots__ = ['_direction', '_pins']

    class Direction(Enum):
        """
        Define the possible directions for a given port

        Possible Directions are:

        UNDEFINED, INOUT, IN, OUT
        """
        UNDEFINED = 0
        INOUT = 1
        IN = 2
        OUT = 3

    def __init__(self):
        """
        setup an empty port
        """
        super().__init__()
        self._direction = self.Direction.UNDEFINED
        self._pins = list()

    def _items(self):
        '''overrides the _items function in the bundles class. For ports, pins are returned'''
        return self._pins

    @property
    def direction(self):
        '''get the direction of the port. This will be a variable of type Port.Direction'''
        return self._direction

    @direction.setter
    def direction(self, value):
        '''set the direction of the port.

        parameters
        ----------

        value - (Port.Direction or int or str) when a Port.Direction is passed in it will set the port accordingly. when an int is passed in it will be 0: UNDEFINED, 1: INOUT, 2: IN, 3: OUT. if a string is passed in it is case insensitively compared with the names and assigned accordingly
        '''
        if isinstance(value, self.Direction):
            self._direction = value
        elif isinstance(value, int):
            for direction in self.Direction:
                if direction.value == value:
                    self._direction = direction
                    break
        elif isinstance(value, str):
            value = value.lower()
            for direction in self.Direction:
                if direction.name.lower() == value:
                    self._direction = direction
                    break
        else:
            raise TypeError("Type {} cannot be assigned to direction".format(type(value)))

    @property
    def pins(self):
        '''get a list of the pins that are in the port'''
        return ListView(self._pins)

    @pins.setter
    def pins(self, value):
        '''this function can set the pins for the port, but it can only be used to reorder the pins in the port.
        It cannot be used to add or remove pins from the port. to do this use the add_pin or remove_pin functions instead

        parameters
        ----------

        value - (List of InnerPin objects) the reordered pins'''
        value_list = list(value)
        value_set = set(value_list)
        assert len(value_set) == len(value_list) and set(self._pins) == value_set, \
            "Set of values do not match, assignment can only be used to reorder values, values must be unique"
        self._pins = value_list

    def create_pins(self, pin_count):
        """
        create pin_count pins in the given port a downto style syntax is assumed

        Parameters
        ----------

        pin_count - (int) this is the number of pins to add to the port
        """
        for _ in range(pin_count):
            self.create_pin()
        return self.pins[-pin_count:]

    def create_pin(self):
        """
        create a pin and add it to the port.
        return:
        the inner_pin created
        """
        pin = InnerPin()
        self.add_pin(pin)
        if self.definition:
            for reference in self.definition.references:
                reference._pins[pin] = OuterPin(reference, pin)
        return pin

    def add_pin(self, pin, position=None):
        """
        add a pin to the port at the given position.

        parameters
        ----------

        pin - (Pin) the pin to be added to the port.

        position - (int, default None) the index at which to add the pin
        """
        assert isinstance(pin, InnerPin)
        assert pin.port is not self, "Pin already belongs to this port"
        assert pin.port is None, "Pin already belongs to another port"
        global_callback._call_port_add_pin(self, pin)
        if position is None:
            self._pins.append(pin)
        else:
            self._pins.insert(position, pin)
        pin._port = self

    def remove_pin(self, pin):
        '''
        remove the given pin from the port. The pin must belong to the port in order to be removed. Wires are disconnected from the pin that is removed.

        parameters
        ----------

        pin - (Pin) a pin to be removed from the port.
        '''
        assert pin.port == self, "Pin does not belong to this port."
        self._remove_pin(pin)
        self._pins.remove(pin)

    def remove_pins_from(self, pins):
        '''
        remove several pins from the port at once. The wires are disconnected from the pins that are removed.

        parameters
        ----------

        pins - (List of Pin objects) a list of all pins to be removed from the port.

        '''
        if isinstance(pins, set):
            exclude_pins = pins
        else:
            exclude_pins = set(pins)
        assert all(isinstance(x, InnerPin) and x.port == self for x in exclude_pins), "All pins to remove must be " \
                                                                                      "InnerPins and belong to the port"
        for pin in exclude_pins:
            self._remove_pin(pin)
        self._pins = list(x for x in self._pins if x not in exclude_pins)

    def _remove_pin(self, pin):
        '''internal pin removal function. disconnects the wires from the pin and remvoes all the pins reference to other pins.'''
        global_callback._call_port_remove_pin(self, pin)
        if self.definition:
            for reference in self.definition.references:
                outer_pin = reference.pins[pin]
                wire = outer_pin.wire
                if wire:
                    wire.disconnect_pin(outer_pin)
                del reference._pins[pin]
                outer_pin._instance = None
                outer_pin._inner_pin = None
        pin._port = None


from spydrnet.ir.element import Element
from spydrnet.ir.outerpin import OuterPin
from spydrnet.ir.views.outerpinsview import OuterPinsView
from spydrnet.global_state import global_callback


class Instance(Element):
    """
    netlist instance of a netlist definition. Instances are literally instances of definitions and they reside inside definitions.
    Function names have been set to adjust for the potential confusion that could arise because instances both have a parent definition and have definitions which they reference.
    """
    __slots__ = ['_parent', '_reference', '_pins']

    def __init__(self):
        """
        creates an empty object of type instance
        """
        super().__init__()
        self._parent = None
        self._reference = None
        self._pins = dict()

    @property
    def parent(self):
        '''Get the definition that contains this instance'''
        return self._parent

    @property
    def reference(self):
        '''get the definition that this instance is instantiating'''
        return self._reference

    @reference.setter
    def reference(self, value):
        '''change the definition that represents this instance.
        Port positioning and size must be taken into account when a new definition is being used.
        if they are different the connections cannot be done automatically with this function.

        parameters
        ----------

        value - (Definition) the definition that this instance should be an instance of'''
        global_callback._call_instance_reference(self, value)
        if value is None:
            for pin in self.pins:
                wire = pin.wire
                if wire:
                    wire.disconnect_pin(pin)
                if isinstance(pin, OuterPin):
                    pin._instance = None
                    pin._inner_pin = None
            self._pins.clear()
            if self._reference:
                self._reference._references.remove(self)
        else:
            if self._reference is not None:
                assert len(self.reference.ports) == len(value.ports) and all(len(x.pins) == len(y.pins) for x, y in
                                                                             zip(self.reference.ports, value.ports)), \
                    "Reference reassignment only supported for definitions with matching port positions"
                self._reference._references.remove(self)
                for cur_port, new_port in zip(self.reference.ports, value.ports):
                    for cur_pin, new_pin in zip(cur_port.pins, new_port.pins):
                        outer_pin = self._pins.pop(cur_pin)
                        outer_pin._inner_pin = new_pin
                        self._pins[new_pin] = outer_pin
            else:
                for port in value.ports:
                    for pin in port.pins:
                        self._pins[pin] = OuterPin.from_instance_and_inner_pin(self, pin)
            value._references.add(self)
        self._reference = value

    @reference.deleter
    def reference(self):
        self.reference = None

    @property
    def pins(self):
        '''get the pins on this instance.'''
        return OuterPinsView(self._pins)


from spydrnet.ir.pin import Pin


class InnerPin(Pin):
    """
    Pins that correspond to definitions. These pins can be thought of as on the inside of a definition. There can be
    many outer pins for each inner pin
    """
    __slots__ = ['_port']

    def __init__(self):
        super().__init__()
        self._port = None

    @property
    def port(self):
        '''return the port that the inner pin is a part of. This object cannot be modified directly by the end user.'''
        return self._port



import unittest

import spydrnet as sdn
from spydrnet.ir.element import Element


class TestDefinition(unittest.TestCase):
    def setUp(self):
        self.definition = sdn.Definition()

    def test_constructor(self):
        self.assertIsInstance(self.definition, Element, "Definition is not an element.")
        self.assertTrue(self.definition, "Constructor returns None type or empty collection.")
        definition2 = sdn.Definition()
        self.assertNotEqual(self.definition, definition2, "Unique objects are considered equal.")

    @unittest.expectedFailure
    def test_assign_library(self):
        library = sdn.Library()
        self.definition.library = library

    def test_ports_set(self):
        port1 = self.definition.create_port()
        port2 = self.definition.create_port()
        self.assertEqual(self.definition.ports, [port1, port2])
        self.definition.ports = [port2, port1]
        self.assertEqual(self.definition.ports, [port2, port1])

    def test_create_port(self):
        port = self.definition.create_port()
        self.assertTrue(port in self.definition.ports)
        self.assertEqual(port.definition, self.definition)

    def test_add_port(self):
        port = sdn.Port()
        self.definition.add_port(port, position=0)
        self.assertTrue(port in self.definition.ports)
        self.assertEqual(port.definition, self.definition)
        self.assertEqual(self.definition.ports.count(port), 1)

    def test_remove_port(self):
        port = self.definition.create_port()
        self.definition.remove_port(port)
        self.assertFalse(port in self.definition.ports)
        self.assertIsNone(port.definition)

    @unittest.expectedFailure
    def test_remove_ports_from_outside_definition(self):
        port = sdn.Port()
        self.definition.remove_ports_from((port,))

    def test_remove_ports_from(self):
        port_included = self.definition.create_port()
        port = self.definition.create_port()
        self.definition.remove_ports_from((port,))
        self.assertFalse(port in self.definition.ports)
        self.assertIsNone(port.definition)

        port = self.definition.create_port()
        self.definition.remove_ports_from({port})
        self.assertFalse(port in self.definition.ports)
        self.assertIsNone(port.definition)

        self.assertTrue(port_included in self.definition.ports)
        self.assertEqual(port_included.definition, self.definition)

    def test_cables_set(self):
        cable1 = self.definition.create_cable()
        cable2 = self.definition.create_cable()
        self.assertEqual(self.definition.cables, [cable1, cable2])
        self.definition.cables = [cable2, cable1]
        self.assertEqual(self.definition.cables, [cable2, cable1])

    def test_create_cable(self):
        cable = self.definition.create_cable()
        self.assertTrue(cable in self.definition.cables)
        self.assertEqual(self.definition, cable.definition)

    def test_add_cable(self):
        cable = sdn.Cable()
        self.definition.add_cable(cable, position=0)
        self.assertTrue(cable in self.definition.cables)
        self.assertEqual(cable.definition, self.definition)
        self.assertEqual(self.definition.cables.count(cable), 1)

    def test_remove_cable(self):
        cable = self.definition.create_cable()
        self.definition.remove_cable(cable)
        self.assertFalse(cable in self.definition.cables)
        self.assertIsNone(cable.definition)

    @unittest.expectedFailure
    def test_remove_cables_from_outside_definition(self):
        cable = sdn.Cable()
        self.definition.remove_cables_from({cable})

    def test_remove_cables_from(self):
        cable_included = self.definition.create_cable()
        cable = self.definition.create_cable()
        self.definition.remove_cables_from({cable})
        self.assertFalse(cable in self.definition.cables)
        self.assertIsNone(cable.definition)
        self.assertTrue(cable_included in self.definition.cables)
        self.assertEqual(cable_included.definition, self.definition)

    def test_children_set(self):
        child1 = self.definition.create_child()
        child2 = self.definition.create_child()
        self.assertEqual(self.definition.children, [child1, child2])
        self.definition.children = [child2, child1]
        self.assertEqual(self.definition.children, [child2, child1])

    def test_create_child(self):
        instance = self.definition.create_child()
        self.assertTrue(instance in self.definition.children)
        self.assertEqual(instance.parent, self.definition)

    def test_add_child(self):
        instance = sdn.Instance()
        self.definition.add_child(instance, position=0)
        self.assertTrue(instance in self.definition.children)
        self.assertEqual(instance.parent, self.definition)
        self.assertEqual(self.definition.children.count(instance), 1)

    def test_remove_child(self):
        instance = self.definition.create_child()
        self.definition.remove_child(instance)
        self.assertFalse(instance in self.definition.children)
        self.assertIsNone(instance.parent)

    @unittest.expectedFailure
    def test_remove_children_from_outside_definition(self):
        instance = sdn.Instance()
        self.definition.remove_children_from((instance,))

    def test_remove_children_from(self):
        instance_included = self.definition.create_child()

        instance = self.definition.create_child()
        self.definition.remove_children_from((instance,))
        self.assertFalse(instance in self.definition.children)
        self.assertIsNone(instance.parent)

        instance = self.definition.create_child()
        self.definition.remove_children_from({instance})
        self.assertFalse(instance in self.definition.children)
        self.assertIsNone(instance.parent)

        self.assertTrue(instance_included in self.definition.children)
        self.assertEqual(instance_included.parent, self.definition)

    def test_is_leaf(self):
        self.assertTrue(self.definition.is_leaf()), "Empty definition is not considered a leaf cell"
        self.definition.create_port()
        self.assertTrue(self.definition.is_leaf()), "Empty definition with a port is not considered a leaf cell"
        self.definition.create_cable()
        self.assertFalse(self.definition.is_leaf()), "Definition with a cable is considered a leaf cell"
        self.definition.remove_cables_from(self.definition.cables)
        self.definition.create_child()
        self.assertFalse(self.definition.is_leaf()), "Definition with a child instance is considered a leaf cell"
        self.definition.create_cable()
        self.assertFalse(self.definition.is_leaf()), "Definition with a cable and child instance is considered a leaf" \
                                                     " cell"


from spydrnet.ir.element import Element
from spydrnet.ir.definition import Definition
from spydrnet.ir.views.listview import ListView
from spydrnet.global_state import global_callback


class Library(Element):
    """
    Represents a library object.

    Contains a pointer to parent netlist and definitions.
    """
    __slots__ = ['_netlist', '_definitions']

    def __init__(self):
        super().__init__()
        self._netlist = None
        self._definitions = list()

    @property
    def netlist(self):
        """
        get the netlist that contains this library
        """
        return self._netlist

    @property
    def definitions(self):
        """
        return a list of all the definitions that are included in this library
        """
        return ListView(self._definitions)

    @definitions.setter
    def definitions(self, value):
        """
        set the definitions to a new reordered set of definitions. This function cannot be used to add or remove
        definitions

        Parameters
        ----------

        value - (List containing Definition type objects) The reordered list
        """
        value_list = list(value)
        value_set = set(value_list)
        assert len(value_list) == len(value_set) and set(self._definitions) == value_set, \
            "Set of values do not match, this function can only reorder values, values must be unique"
        self._definitions = value_list

    def create_definition(self):
        """
        create a definition, add it to the library, and return the definition
        """
        definition = Definition()
        self.add_definition(definition)
        return definition

    def add_definition(self, definition, position=None):
        """
        add an existing definition to the library. The definition must not belong to a library including this one.

        parameters
        ----------

        definition - (Definition) the defintion to add to the library

        position - (int, default None) the index in the library list at which to add the definition

        """
        assert definition.library is not self, "Definition already included in library"
        assert definition.library is None, "Definition already belongs to a different library"
        global_callback._call_library_add_definition(self, definition)
        if position is not None:
            self._definitions.insert(position, definition)
        else:
            self._definitions.append(definition)
        definition._library = self

    def remove_definition(self, definition):
        """
        remove the given definition from the library

        parameters
        ----------

        definition - (Definition) the definition to be removed
        """
        assert definition.library == self, "definition is not included in library"
        self._remove_definition(definition)
        self._definitions.remove(definition)

    def remove_definitions_from(self, definitions):
        """
        remove a set of definitions from the library. all definitions provided must be in the library

        parameters
        ----------

        definitions - (Set of Definition type objects) the definitions to be removed
        """
        if isinstance(definitions, set):
            excluded_definitions = definitions
        else:
            excluded_definitions = set(definitions)
        assert all(x.library == self for x in excluded_definitions), "Some definitions to remove are not included in " \
                                                                     "the library "
        included_definitions = list()
        for definition in self._definitions:
            if definition not in excluded_definitions:
                included_definitions.append(definition)
            else:
                self._remove_definition(definition)
        self._definitions = included_definitions

    def _remove_definition(self, definition):
        """
        internal function to dissociate a definition from the library
        """
        global_callback._call_library_remove_definition(self, definition)
        definition._library = None


import unittest

import spydrnet as sdn
from spydrnet.ir.bundle import Bundle
from spydrnet.ir.element import Element


class TestBundle(unittest.TestCase):
    def setUp(self) -> None:
        self.bundle = Bundle()

    def test_constructor(self):
        self.assertIsInstance(self.bundle, Element, "Bundle is not an element.")
        self.assertTrue(self.bundle, "Constructor returns None type or empty collection.")
        bundle2 = Bundle()
        self.assertNotEqual(self.bundle, bundle2, "Unique objects are considered equal.")

    def test_definition(self):
        self.assertIsNone(self.bundle.definition)

    @unittest.expectedFailure
    def test_definition_assignment(self):
        definition = sdn.Definition()
        self.bundle.definition = definition

    def test_isdownto(self):
        self.assertTrue(self.bundle.is_downto)
        self.bundle.is_downto = False
        self.assertFalse(self.bundle.is_downto)

    def test_lower_index(self):
        self.assertEqual(self.bundle.lower_index, 0)
        self.bundle.lower_index = 1
        self.assertEqual(self.bundle.lower_index, 1)

    def test__item(self):
        self.assertIsNone(self.bundle._items())


import unittest

from spydrnet.ir.views.setview import SetView


class TestSetView(unittest.TestCase):
    def setUp(self) -> None:
        self.set_view = SetView(set(range(10)))

    def test_and(self):
        self.assertEqual(len(set(range(10, 20)) & self.set_view), 0)
        self.assertEqual(len(self.set_view & set(range(10, 20))), 0)

    def test_or(self):
        self.assertEqual(len(set(range(10, 20)) | self.set_view), 20)
        self.assertEqual(len(self.set_view | set(range(10, 20))), 20)

    def test_min(self):
        self.assertEqual(len(set(range(10, 20)) - self.set_view), 10)
        self.assertEqual(len(self.set_view - set(range(10, 20))), 10)

    def test_xor(self):
        self.assertEqual(len(set(range(10, 20)) ^ self.set_view), 20)
        self.assertEqual(len(self.set_view ^ set(range(10, 20))), 20)

    @unittest.expectedFailure
    def test_and_assignment(self):
        self.set_view &= set(range(5))

    @unittest.expectedFailure
    def test_or_assignment(self):
        self.set_view |= set(range(5))

    @unittest.expectedFailure
    def test_min_assignment(self):
        self.set_view -= set(range(5))

    @unittest.expectedFailure
    def test_xor_assignment(self):
        self.set_view ^= set(range(5))

    def test_hash(self):
        self.assertRaises(TypeError, hash, self.set_view)

    def test_eq(self):
        self.assertTrue(self.set_view == set(range(10)))
        self.assertTrue(self.set_view == SetView(set(range(10))))

    def test_ne(self):
        self.assertFalse(self.set_view != set(range(10)))
        self.assertFalse(self.set_view != SetView(set(range(10))))

    def test_ge(self):
        self.assertTrue(self.set_view >= set(range(10)))
        self.assertTrue(self.set_view >= SetView(set(range(10))))

    def test_le(self):
        self.assertTrue(self.set_view <= set(range(10)))
        self.assertTrue(self.set_view <= SetView(set(range(10))))

    def test_gt(self):
        self.assertFalse(self.set_view > set(range(10)))
        self.assertFalse(self.set_view > SetView(set(range(10))))

    def test_lt(self):
        self.assertFalse(self.set_view < set(range(10)))
        self.assertFalse(self.set_view < SetView(set(range(10))))

    def test_len(self):
        self.assertTrue(len(self.set_view), 10)

    def test_repr(self):
        self.assertEqual(repr(self.set_view), repr(set(range(10))))

    def test_str(self):
        self.assertEqual(str(self.set_view), str(set(range(10))))

    def test_copy(self):
        self.assertEqual(set(range(10)), self.set_view.copy())

    def test_difference(self):
        self.assertEqual(set(range(5)), self.set_view.difference(set(range(5, 10))))

    def test_intersection(self):
        self.assertEqual(set(range(5)), self.set_view.intersection(set(range(5))))

    def test_isdisjoint(self):
        self.assertTrue(self.set_view.isdisjoint(set(range(10, 20))))

    def test_issubset(self):
        self.assertTrue(self.set_view.issubset(set(range(20))))

    def test_issuperset(self):
        self.assertTrue(self.set_view.issuperset(set(range(5))))

    def test_symmetric_difference(self):
        self.assertTrue(set(range(10)) ^ set(range(5, 20)), self.set_view.symmetric_difference(set(range(5, 20))))

    def test_union(self):
        self.assertTrue(set(range(20)), self.set_view.union(set(range(10, 20))))




from spydrnet.ir.views.dictview import DictView
from spydrnet.ir.outerpin import OuterPin


class OuterPinsView(DictView):

    def __init__(self, dict_object):
        super().__init__(dict_object)

    def __contains__(self, item):
        if item not in self._dict:
            if isinstance(item, OuterPin):
                return item.inner_pin in self._dict
            return False
        return True

    def __getitem__(self, item):
        if isinstance(item, OuterPin):
            return self._dict[item.inner_pin]
        return self._dict[item]

    def __iter__(self):
        return iter(self._dict.values())

    def get(self, key, default=None):
        if isinstance(key, OuterPin):
            return self._dict.get(key.inner_pin, default)
        return self._dict.get(key, default)



class DictView:
    __slots__ = ['_dict']

    def __init__(self, dict_object):
        self._dict = dict_object

    def __eq__(self, other):
        if isinstance(other, DictView):
            return self._dict.__eq__(other._dict)
        return self._dict.__eq__(other)

    def __hash__(self):
        return self._dict.__hash__()

    def __len__(self):
        return self._dict.__len__()

    def __ne__(self, other):
        if isinstance(other, DictView):
            return self._dict.__ne__(other._dict)
        return self._dict.__ne__(other)

    def __repr__(self):
        return self._dict.__repr__()

    def __str__(self):
        return self._dict.__str__()

    def __contains__(self, item):
        return self._dict.__contains__(item)

    def __getitem__(self, item):
        return self._dict.__getitem__(item)

    def __iter__(self):
        return self._dict.__iter__()

    def get(self, key, default=None):
        return self._dict.get(key, default)

    def copy(self):
        return self._dict.copy()

    def fromkeys(self, *args, **kwargs):
        return self._dict.fromkeys(*args, **kwargs)

    def items(self):
        return self._dict.items()

    def keys(self):
        return self._dict.keys()

    def values(self):
        return self._dict.values()



from spydrnet.ir.bundle import Bundle
from spydrnet.ir.wire import Wire
from spydrnet.ir.views.listview import ListView
from spydrnet.global_state import global_callback


class Cable(Bundle):
    '''Much like Ports cable extend the bundle class, giving them indexing ability they represent several wires in a collection or bus that are generally related.
    This could be thought of much like vector types in VHDL ie std_logic_vector (7 downto 0)'''
    __slots__ = ['_wires']

    def __init__(self):
        '''create a cable with no wires and default values for a bundle.'''
        super().__init__()
        self._wires = list()

    def _items(self):
        '''overrides the bundle _items function to return wires'''
        return self._wires

    @property
    def wires(self):
        '''get a list of wires that are in this cable'''
        return ListView(self._wires)

    @wires.setter
    def wires(self, value):
        '''set the wires to a reordered list of wires. This function is to be used for reordering of wires'''
        value_list = list(value)
        value_set = set(value_list)
        assert len(value_list) == len(value_set) and set(self._wires) == value_set, \
            "Set of values does not match, assigment can only be used for reordering, values must be unique"
        self._wires = value_list

    def create_wires(self, wire_count):
        '''creates wire_count wires for this cable and adds them to it.

        parameters
        ----------

        wire_count - (int) the number of wires to be added to the cable.'''
        for _ in range(wire_count):
            self.create_wire()
        return self.wires[-wire_count:]

    def create_wire(self):
        '''creates a wire and adds it to the cable. returns the wire that was created'''
        wire = Wire()
        self.add_wire(wire)
        return wire

    def add_wire(self, wire, position=None):
        '''adds a wire to the cable at the given position. This wire must not belong to a cable already

        parameters
        ----------

        wire - (Wire) the wire to be added to the cable. This wire must not belong to any other cable.

        position - (int, default None) the index in the wires list at which to add the wire.'''
        assert wire.cable is not self, "Wire already belongs to this cable"
        assert wire.cable is None, "Wire already belongs to a different cable"
        global_callback._call_cable_add_wire(self, wire)
        if position is not None:
            self._wires.insert(position, wire)
        else:
            self._wires.append(wire)
        wire._cable = self

    def remove_wire(self, wire):
        '''remove the given wire from the cable and return it. The wire must belong to this cable

        parameters
        ----------

        wire - (Wire) the wire to be removed from the cable.'''
        assert wire.cable == self, "Wire does not belong to this cable"
        self._remove_wire(wire)
        self._wires.remove(wire)

    def remove_wires_from(self, wires):
        '''remove all wires given from the cable. Each must be a member of this cable.

        parameters
        ----------

        wires - (List of Wire objects) wires to be removed from the cable.'''
        if isinstance(wires, set):
            excluded_wires = wires
        else:
            excluded_wires = set(wires)
        assert all(x.cable == self for x in excluded_wires), "Some wires do not belong to this cable"
        for wire in excluded_wires:
            self._remove_wire(wire)
        self._wires = list(x for x in self._wires if x not in excluded_wires)

    def _remove_wire(self, wire):
        '''internal wire removal call. dissociates the wire from the cable'''
        global_callback._call_cable_remove_wire(self, wire)
        wire._cable = None



import unittest
import spydrnet as sdn

from spydrnet.plugins.data_manager import DataManager
from spydrnet.global_state.global_netlist import set_current_netlist, get_current_netlist


class TestDataManager(unittest.TestCase):
    def gen_netlist(self):
        netlist = sdn.Netlist()
        set_current_netlist(netlist)
        return netlist

    def gen_library(self):
        netlist = self.gen_netlist()
        lib = netlist.create_library()
        return lib

    def gen_definition(self):
        lib = self.gen_library()
        defin = lib.create_definition()
        return defin
        

    def test_basic_setup(self):
        netlist = self.gen_netlist()
        dm = DataManager()
        lib1 = netlist.create_library()
        lib2 = netlist.create_library()
        lib1['EDIF.identifier'] = "my_lib1"
        lib2['EDIF.identifier'] = "my_lib2"
        def1 = lib1.create_definition()
        def1['EDIF.identifier'] = "d1"
        def2 = lib2.create_definition()
        def2['EDIF.identifier'] = "d1"
        def3 = lib1.create_definition()
        def3['EDIF.identifier'] = "my_lib1"
        c1 = def1.create_cable()
        p1 = def1.create_port()
        i1 = def1.create_child()
        c2 = def1.create_cable()
        p2 = def1.create_port()
        i2 = def1.create_child()
        c1['EDIF.identifier'] = "1"
        i1['EDIF.identifier'] = "1"
        p1['EDIF.identifier'] = "1"
        c2['EDIF.identifier'] = "2"
        i2['EDIF.identifier'] = "2"
        p2['EDIF.identifier'] = "2"

        

    def test_dont_track_orphaned(self):
        netlist = self.gen_netlist()
        dm = DataManager()
        lib1 = sdn.Library()
        lib2 = sdn.Library()
        lib1['EDIF.identifier'] = "my_lib1"
        lib2['EDIF.identifier'] = "my_lib1"

    @unittest.expectedFailure
    def test_duplicate_library_name(self):
        netlist = self.gen_netlist()
        dm = DataManager()
        lib1 = netlist.create_library()
        lib2 = netlist.create_library()
        lib1['EDIF.identifier'] = "my_lib"
        lib2['EDIF.identifier'] = "my_lib"

    
    @unittest.expectedFailure
    def test_duplicate_definition_name(self):
        lib1 = self.gen_library()
        dm = DataManager()
        def1 = lib1.create_definition()
        def2 = lib1.create_definition()
        def1['EDIF.identifier'] = "my_lib"
        def2['EDIF.identifier'] = "my_lib"

    def test_duplicate_definition_elements(self):
        def1 = self.gen_definition()
        dm = DataManager()
        port = def1.create_port()
        instance = def1.create_child()
        cable = def1.create_cable()
        port['EDIF.identifier'] = "my_lib"
        instance['EDIF.identifier'] = "my_lib"
        cable['EDIF.identifier'] = "my_lib"

    @unittest.expectedFailure
    def test_duplicate_definition_ports(self):
        def1 = self.gen_definition()
        dm = DataManager()
        port = def1.create_port()
        port2 = def1.create_port()
        port['EDIF.identifier'] = "my_lib"
        port2['EDIF.identifier'] = "my_lib"

    @unittest.expectedFailure
    def test_duplicate_definition_cables(self):
        def1 = self.gen_definition()
        dm = DataManager()
        cable = def1.create_cable()
        cable2 = def1.create_cable()
        cable['EDIF.identifier'] = "my_lib"
        cable2['EDIF.identifier'] = "my_lib"

    @unittest.expectedFailure
    def test_duplicate_definition_children(self):
        def1 = self.gen_definition()
        dm = DataManager()
        instance = def1.create_child()
        instance2 = def1.create_child()
        instance['EDIF.identifier'] = "my_lib"
        instance2['EDIF.identifier'] = "my_lib"

    def test_rename(self):
        netlist = self.gen_netlist()
        dm = DataManager()
        lib1 = netlist.create_library()
        lib1['EDIF.identifier'] = "my_lib1"
        lib1['EDIF.identifier'] = "my_lib2"
        lib1['EDIF.identifier'] = "my_lib1"
        lib2 = netlist.create_library()
        lib2['EDIF.identifier'] = "my_lib2"
        def1 = lib1.create_definition()
        def1['EDIF.identifier'] = "my_lib1"
        def1['EDIF.identifier'] = "my_lib2"
        def1['EDIF.identifier'] = "my_lib1"
        def2 = lib1.create_definition()
        def2['EDIF.identifier'] = "my_lib2"
        c = def1.create_cable()
        c['EDIF.identifier'] = "1"
        c['EDIF.identifier'] = "2"
        c['EDIF.identifier'] = "1"
        p = def1.create_port()
        p['EDIF.identifier'] = "1"
        p['EDIF.identifier'] = "2"
        p['EDIF.identifier'] = "1"
        i = def1.create_child()
        i['EDIF.identifier'] = "1"
        i['EDIF.identifier'] = "2"
        i['EDIF.identifier'] = "1"

    def test_remove(self):
        netlist = self.gen_netlist()
        dm = DataManager()
        lib1 = netlist.create_library()
        lib1['EDIF.identifier'] = "my_lib1"
        netlist.remove_library(lib1)
        lib2 = netlist.create_library()
        lib2['EDIF.identifier'] = "my_lib1"
        def1 = lib2.create_definition()
        def1['EDIF.identifier'] = "my_lib1"
        lib2.remove_definition(def1)
        def2 = lib2.create_definition()
        def2['EDIF.identifier'] = "my_lib1"
        c1 = def2.create_cable()
        c2 = def2.create_cable()
        p1 = def2.create_port()
        p2 = def2.create_port()
        i1 = def2.create_child()
        i2 = def2.create_child()
        c1['EDIF.identifier'] = "1"
        def2.remove_cable(c1)
        c2['EDIF.identifier'] = "1"
        p1['EDIF.identifier'] = "1"
        def2.remove_port(p1)
        p2['EDIF.identifier'] = "1"
        i1['EDIF.identifier'] = "1"
        def2.remove_child(i1)
        i2['EDIF.identifier'] = "1"

    def test_orphaned_add(self):
        netlist = self.gen_netlist()
        dm = DataManager()
        lib1 = sdn.Library()
        lib1["EDIF.identifier"] = '1'
        netlist.add_library(lib1)

    @unittest.expectedFailure
    def test_orphaned_add_collision(self):
        netlist = self.gen_netlist()
        dm = DataManager()
        lib1 = sdn.Library()
        lib1["EDIF.identifier"] = '1'
        netlist.add_library(lib1)
        lib2 = sdn.Library()
        lib2["EDIF.identifier"] = '1'
        netlist.add_library(lib2)

    @unittest.expectedFailure
    def test_remove_twice_library(self):
        netlist = self.gen_netlist()
        dm = DataManager()
        lib1 = netlist.create_library()
        lib1['EDIF.identifier'] = "my_lib1"
        netlist.remove_library(lib1)
        netlist.remove_library(lib1)

    @unittest.expectedFailure
    def test_remove_twice_definition(self):
        lib = self.gen_library()
        dm = DataManager()
        d1 = lib.create_definition()
        d1['EDIF.identifier'] = "1"
        lib.remove_definition(d1)
        lib.remove_definition(d1)

    def test_remove_untracked(self):
        netlist = self.gen_netlist()
        dm = DataManager()
        lib1 = netlist.create_library()
        def1 = lib1.create_definition()
        c1 = def1.create_cable()
        p1 = def1.create_port()
        i1 = def1.create_child()
        def1.remove_cable(c1)
        def1.remove_child(i1)
        def1.remove_port(p1)
        lib1.remove_definition(def1)
        netlist.remove_library(lib1)

    def test_remove_tracked(self):
        netlist = self.gen_netlist()
        dm = DataManager()
        lib1 = netlist.create_library()
        lib1["EDIF.identifier"] = "test"
        def1 = lib1.create_definition()
        def1["EDIF.identifier"] = "test"
        c1 = def1.create_cable()
        c1["EDIF.identifier"] = "test"
        p1 = def1.create_port()
        p1["EDIF.identifier"] = "test"
        i1 = def1.create_child()
        i1["EDIF.identifier"] = "test"
        def1.remove_cable(c1)
        def1.remove_child(i1)
        def1.remove_port(p1)
        lib1.remove_definition(def1)
        netlist.remove_library(lib1)

    def test_pop_name(self):
        netlist = self.gen_netlist()
        dm = DataManager()
        lib1 = netlist.create_library()
        lib1['EDIF.identifier'] = "my_lib1"
        lib1.pop('EDIF.identifier')
        lib2 = netlist.create_library()
        lib2['EDIF.identifier'] = "my_lib1"
        def1 = lib2.create_definition()
        def1['EDIF.identifier'] = "my_lib1"
        def1.pop('EDIF.identifier')
        def2 = lib2.create_definition()
        def2['EDIF.identifier'] = "my_lib1"
        c1 = def2.create_cable()
        c2 = def2.create_cable()
        p1 = def2.create_port()
        p2 = def2.create_port()
        i1 = def2.create_child()
        i2 = def2.create_child()
        c1['EDIF.identifier'] = "1"
        c1.pop('EDIF.identifier')
        c2['EDIF.identifier'] = "1"
        p1['EDIF.identifier'] = "1"
        p1.pop('EDIF.identifier')
        p2['EDIF.identifier'] = "1"
        i1['EDIF.identifier'] = "1"
        i1.pop('EDIF.identifier')
        i2['EDIF.identifier'] = "1"


    '''tests TODO:
    rename an object
    orphan an object and see what happens
    ...
    '''


from spydrnet.ir.element import Element
from spydrnet.ir.library import Library
from spydrnet.ir.instance import Instance
from spydrnet.ir.views.listview import ListView
from spydrnet.global_state import global_callback


class Netlist(Element):
    """
    Represents a netlist object.

    Contains a top level instance and libraries
    """
    __slots__ = ['_libraries', '_top_instance']

    def __init__(self):
        super().__init__()
        self._libraries = list()
        self._top_instance = None

    @property
    def libraries(self):
        """get a list of all libraries included in the netlist"""
        return ListView(self._libraries)

    @libraries.setter
    def libraries(self, value):
        """
        set the libraries. This function can only be used to reorder the libraries. Use the remove_library and
        add_library functions to add and remove libraries.

        parameters
        ----------

        value - the reordered list of libraries
        """
        value_list = list(value)
        value_set = set(value_list)
        assert len(value_list) == len(value_set) and set(self._libraries) == value_set, \
            "Set of values do not match, this assignment can only reorder values, values must be unique"
        self._libraries = value_list

    @property
    def top_instance(self):
        """
        Get the top instance in the netlist.

        Returns
        -------
        Instance
            The top level instance in the environment
        """
        return self._top_instance

    @top_instance.setter
    def top_instance(self, instance):
        """
        sets the top instance of the design. The instance must not be null and should probably come from this netlist

        parameters
        ----------

        instance - (Instance) the instance to set as the top instance.
        """
        assert instance is None or isinstance(instance, Instance), "Must specify an instance"
        global_callback._call_netlist_top_instance(self, instance)
        # TODO: should We have a DRC that makes sure the instance is of a definition contained in netlist? I think no
        #  but I am open to hear other points of veiw.
        self._top_instance = instance

    def create_library(self):
        '''create a library and add it to the netlist and return that library'''
        library = Library()
        self.add_library(library)
        return library

    def add_library(self, library, position=None):
        """
        add an already existing library to the netlist. This library should not belong to another netlist. Use
        remove_library from other netlists before adding

        parameters
        ----------

        library - (Library) the library to be added to the netlist

        position - (int, default None) when set it is the index at which to add the library in the libraries list

        """
        assert library not in self._libraries, "Library already included in netlist"
        assert library.netlist is None, "Library already belongs to a different netlist"
        global_callback._call_netlist_add_library(self, library)
        if position is not None:
            self._libraries.insert(position, library)
        else:
            self._libraries.append(library)
        library._netlist = self

    def remove_library(self, library):
        """
        removes the given library if it is in the netlist

        parameters
        ----------

        library - (Library) the library to be removed
        """
        assert library.netlist == self, "Library is not included in netlist"
        self._remove_library(library)
        self._libraries.remove(library)

    def remove_libraries_from(self, libraries):
        '''removes all the given libraries from the netlist. All libraries must be in the netlist

        parameters
        ----------

        libraries - (Set) libraries to be removed
        '''
        if isinstance(libraries, set):
            excluded_libraries = libraries
        else:
            excluded_libraries = set(libraries)
        assert all(x.netlist == self for x in excluded_libraries), "Some libraries to remove are not included in " \
                                                                   "netlist "
        included_libraries = list()
        for library in self._libraries:
            if library not in excluded_libraries:
                included_libraries.append(library)
            else:
                self._remove_library(library)
        self._libraries = included_libraries

    def _remove_library(self, library):
        """
        internal function which will separate a particular libraries binding from the netlist
        """
        global_callback._call_netlist_remove_library(self, library)
        library._netlist = None



import sys
from spydrnet.ir.views.dictview import DictView
from spydrnet.global_state import global_callback


class Element(object):
    """
    Base class of all intermediate representation objects.

    An intermediate representation object represents an item in a netlist. Items range in specificity from pins on a
    port or wires in a cable up to an item that represents the netlist as a whole.

    Each element implements a dictionary for storing key-value pairs. The key should be a case sensitive string and the
    value should be a primitive type (string, integer, float, boolean) or potentially nested collections of primitive
    types. The purpose of this dictionary is to provide a space for properties and metadata associated with the element.

    Key namespaces are separated with a *period* character. If the key is void of a *period* than the key resides in the
    root namespace. Keys in the root namespace are considered properties. Other keys are considered metadata. For
    example '<LANG_OF_ORIGIN>.<METADATA_TAG>':<metadata_value> is considered metadata associated with the netlist's
    language of origin.

    Only data pertinent to the netlist should be stored in this dictionary. Cached data (namespace management, anything
    that can be recreated from the netlist) should be excluded from this dictionary. The intent of the IR is to house
    the basis of data for the netlist.

    The only key that is reserved is 'NAME'. It is the primary name of the element. NAME may be undefined or inferred,
    for example, a pin on a port may be nameless, but infer its name for its parent port and position.
    """
    __slots__ = ['_data']

    def __init__(self):
        """
        Initialize an element with an empty data dictionary.
        """
        self._data = dict()

    @property
    def data(self):
        return DictView(self._data)
        
    @property
    def name(self):
        return self._data.get(".NAME", None)
        
    @name.setter
    def name(self, value):
        if value is None and ".NAME" in self:
            del self[".NAME"]
        else:
            self[".NAME"] = value

    @name.deleter
    def name(self):
        if ".NAME" in self:
            del self[".NAME"]

    def __setitem__(self, key, value):
        """
        create an entry in the dictionary of the element it will be stored in the metadata.
        """
        global_callback._call_dictionary_set(self, key, value)
        key = sys.intern(key)
        self._data.__setitem__(sys.intern(key), value)

    def __delitem__(self, key):
        global_callback._call_dictionary_delete(self, key)
        self._data.__delitem__(key)

    def __getitem__(self, key):
        return self._data.__getitem__(key)

    def __contains__(self, item):
        return self._data.__contains__(item)

    def __iter__(self):
        return self._data.__iter__()

    def pop(self, item):
        global_callback._call_dictionary_pop(self, item)
        return self._data.pop(item)



import unittest

from spydrnet.ir.views.listview import ListView


class TestListView(unittest.TestCase):
    def setUp(self) -> None:
        self.list_view = ListView(list(range(10)))

    def test_getitem(self):
        self.assertTrue(self.list_view[0] == 0)

    def test_contains(self):
        self.assertTrue(5 in self.list_view)

    def test_comparison(self):
        list_copy = list(range(10))
        self.assertEqual(self.list_view, list_copy)
        empty_list = list()
        self.assertNotEqual(self.list_view, empty_list)

    def test_iter(self):
        aggregate = 0
        for item in self.list_view:
            aggregate += item
        self.assertEqual(sum(range(10)), aggregate)

    def test_len(self):
        self.assertEqual(len(self.list_view), 10)

    def test_reversed(self):
        self.assertEqual(next(iter(reversed(self.list_view))), 9)

    def test_copy(self):
        self.assertEqual(self.list_view, self.list_view.copy())

    def test_count(self):
        self.assertEqual(self.list_view.count(5), 1)

    def test_index(self):
        self.assertEqual(self.list_view.index(5), 5)

    def test_slice(self):
        self.assertEqual(self.list_view[1:4], list(range(1, 4)))

    def test_multiply(self):
        self.assertEqual(len(self.list_view * 2), 20)
        self.assertEqual(len(2 * self.list_view), 20)

    def test_add(self):
        result = list(range(10, 20)) + self.list_view
        print(result)
        self.assertEqual(len(result), 20)
        result = self.list_view + list(range(10, 20))
        print(result)
        self.assertEqual(len(result), 20)

    @unittest.expectedFailure
    def test_multiply_assignment(self):
        self.list_view *= 2

    @unittest.expectedFailure
    def test_add_assignment(self):
        self.list_view += list(range(10))

    def test_ge(self):
        self.assertTrue(self.list_view >= ListView(list(range(10))))
        self.assertTrue(self.list_view >= list(range(10)))

    def test_gt(self):
        self.assertTrue(self.list_view > ListView(list(range(9))))
        self.assertTrue(self.list_view > list(range(9)))

    def test_le(self):
        self.assertTrue(self.list_view <= ListView(list(range(10))))
        self.assertTrue(self.list_view <= list(range(10)))

    def test_lt(self):
        self.assertTrue(self.list_view < ListView(list(range(11))))
        self.assertTrue(self.list_view < list(range(11)))

    def test_repr(self):
        self.assertEqual(repr(self.list_view), repr(list(range(10))))

    def test_str(self):
        self.assertEqual(str(self.list_view), str(list(range(10))))

    def test_hash(self):
        self.assertRaises(TypeError, hash, self.list_view)



from spydrnet.ir.outerpin import OuterPin
from spydrnet.ir.views.listview import ListView
from spydrnet.global_state import global_callback


class Wire:
    __slots__ = ['_cable', '_pins']

    def __init__(self):
        self._cable = None
        self._pins = list()

    @property
    def cable(self):
        return self._cable

    @property
    def pins(self):
        return ListView(self._pins)

    @pins.setter
    def pins(self, value):
        value_list = list(value)
        value_set = set(value_list)
        assert len(value_list) == len(value_set) and set(self._pins) == value_set, \
            "Set of values do not match, assignment can only be used to reorder, values must be unique"
        self._pins = value_list

    def connect_pin(self, pin, position=None):
        global_callback._call_wire_connect_pin(self, pin)
        if isinstance(pin, OuterPin):
            instance = pin.instance
            inner_pin = pin.inner_pin
            assert instance is not None and inner_pin is not None, \
                "Outer pin must represent an instance and an inner_pin"
            assert inner_pin in instance.pins, "Pin not associated with instance"
            outer_pin = instance.pins[inner_pin]
            assert outer_pin.wire is not self, "Pin already connected to this wire"
            assert outer_pin.wire is None, "Pin already connected to a different wire"
            pin._wire = self
            pin = outer_pin
        else:
            assert pin.wire is None, "Pin already connected to a different wire"
        if position is not None:
            self._pins.insert(position, pin)
        else:
            self._pins.append(pin)
        pin._wire = self

    def disconnect_pin(self, pin):
        if isinstance(pin, OuterPin):
            instance = pin.instance
            inner_pin = pin.inner_pin
            assert instance is not None and inner_pin is not None, \
                "Outer pin must represent an instance and an inner_pin"
            assert inner_pin in instance.pins, "Pin not associated with instance"
            outer_pin = instance.pins[inner_pin]
            assert outer_pin.wire is self, "Pin is disconnected or connected to a different wire."
            self._disconnect_pin(pin)
            pin = outer_pin
        else:
            assert pin.wire == self, "Pin does not belong to this wire"
        self._pins.remove(pin)
        self._disconnect_pin(pin)

    def disconnect_pins_from(self, pins):
        if isinstance(pins, set):
            excluded_pins = pins
        else:
            excluded_pins = set(pins)
        all_pins_can_be_disconnected = True
        for pin in excluded_pins:
            if isinstance(pin, OuterPin):
                instance = pin.instance
                inner_pin = pin.inner_pin
                if instance is None or inner_pin is None or inner_pin not in instance.pins or \
                        instance.pins[inner_pin].wire is not self:
                    all_pins_can_be_disconnected = False
                    break
            else:
                if pin.wire != self:
                    all_pins_can_be_disconnected = False
                    break
        assert all_pins_can_be_disconnected, "Some of the pins to disconnect are not associated with an instance, " \
                                             "already disconnected, or connected to a different wire"
        for pin in excluded_pins:
            if isinstance(pin, OuterPin):
                self._disconnect_pin(pin)
                pin = pin.instance.pins[pin]
            self._disconnect_pin(pin)
        self._pins = list(x for x in self._pins if x not in excluded_pins)

    def _disconnect_pin(self, pin):
        global_callback._call_wire_disconnect_pin(self, pin)
        pin._wire = None


import unittest

import spydrnet as sdn
from spydrnet.ir.bundle import Bundle


class TestPort(unittest.TestCase):
    def setUp(self) -> None:
        self.port = sdn.Port()

    def test_direction_enum(self):
        self.assertEqual(sdn.Port.Direction.UNDEFINED, sdn.UNDEFINED)
        self.assertEqual(sdn.Port.Direction.IN, sdn.IN)
        self.assertEqual(sdn.Port.Direction.OUT, sdn.OUT)
        self.assertEqual(sdn.Port.Direction.INOUT, sdn.INOUT)

    def test_constructor(self):
        self.assertIsInstance(self.port, Bundle)
        self.assertTrue(self.port, "Constructor returns None type or empty collection.")
        port2 = sdn.Port()
        self.assertNotEqual(self.port, port2, "Unique objects are considered equal.")

    def test_pins_set(self):
        pin1 = self.port.create_pin()
        pin2 = self.port.create_pin()
        self.assertEqual(self.port.pins, [pin1, pin2])
        self.port.pins = [pin2, pin1]
        self.assertEqual(self.port.pins, [pin2, pin1])

    def test_direction(self):
        for ii in range(4):
            self.port.direction = ii
            self.assertEqual(self.port.direction.value, ii)
        directions = ['undefined', 'in', 'out', 'inout']
        for direction in directions:
            self.port.direction = direction
            self.assertEqual(self.port.direction.name.lower(), direction.lower())
        for direction in sdn.Port.Direction:
            self.port.direction = direction
            self.assertEqual(self.port.direction, direction)

    @unittest.expectedFailure
    def test_direction_set_bad_type(self):
        self.port.direction = list()

    def test_initialize_pins(self):
        self.port.create_pins(2)
        self.assertEqual(len(self.port.pins), 2)
        self.assertNotEqual(self.port.pins[0], self.port.pins[1])

    def test_create_pin(self):
        pin = self.port.create_pin()
        self.assertTrue(pin in self.port.pins)
        self.assertEqual(pin.port, self.port)

    def test_add_pin(self):
        pin = sdn.InnerPin()
        self.port.add_pin(pin, position=0)
        self.assertTrue(pin in self.port.pins)
        self.assertEqual(pin.port, self.port)
        self.assertEqual(self.port.pins.count(pin), 1)

    def test_remove_pin(self):
        pin = self.port.create_pin()
        self.port.remove_pin(pin)
        self.assertFalse(pin in self.port)
        self.assertIsNone(pin.port)

    @unittest.expectedFailure
    def test_remove_pins_outside_port(self):
        pin = sdn.InnerPin()
        self.port.remove_pins_from((pin,))

    def test_remove_pins_from(self):
        pin_included = self.port.create_pin()
        pin = self.port.create_pin()
        self.port.remove_pins_from({pin})
        self.assertFalse(pin in self.port.pins)
        self.assertIsNone(pin.port)
        self.assertTrue(pin_included in self.port.pins)
        self.assertEqual(pin_included.port, self.port)

    def test_is_scalar(self):
        self.assertTrue(self.port.is_scalar)
        self.port.is_scalar = False
        self.assertFalse(self.port.is_scalar)
        self.port.is_scalar = True
        self.assertTrue(self.port.is_scalar)
        self.port.create_pins(2)
        self.assertFalse(self.port.is_scalar)
        self.port.is_scalar = False
        self.port.remove_pins_from(self.port.pins)
        self.assertFalse(self.port.is_scalar)

    @unittest.expectedFailure
    def test_is_scalar_set_on_array_bundle(self):
        self.port.create_pins(2)
        self.port.is_scalar = True

    def test_is_array(self):
        self.assertFalse(self.port.is_array)
        self.port.is_array = True
        self.assertTrue(self.port.is_array)
        self.port.is_array = False
        self.assertFalse(self.port.is_array)
        self.port.create_pins(2)
        self.assertTrue(self.port.is_array)
        self.port.is_array = True
        self.port.remove_pins_from(self.port.pins)
        self.assertTrue(self.port.is_array)

    @unittest.expectedFailure
    def test_is_array_clear_on_array_bundle(self):
        self.port.create_pins(2)
        self.port.is_array = False


import unittest

from spydrnet.ir.element import Element


class TestElement(unittest.TestCase):
    def setUp(self) -> None:
        self.element = Element()

    def test_constructor(self):
        element1 = Element()
        self.assertTrue(element1, "Constructor return None type or empty collection")
        element2 = Element()
        self.assertNotEqual(element1, element2, "Unique objects are considered equal.")

    def test_dictionary(self):
        self.assertFalse('NAME' in self.element)
        self.element['NAME'] = "TestName"
        self.assertTrue('NAME' in self.element)
        for key in self.element:
            self.assertEqual(self.element[key], "TestName")
        del self.element['NAME']
        self.assertFalse('NAME' in self.element)
        self.element['NAME'] = "DifferentName"
        name = self.element.pop('NAME')
        self.assertEqual(name, "DifferentName")

    def test_name(self):
        self.element.name = "TestName"
        self.assertTrue(".NAME" in self.element)
        self.assertEqual(self.element.name, "TestName")

    def test_del_name(self):
        self.assertIsNone(self.element.name)
        self.element.name = None
        self.element.name = "TestName"
        self.element.name = None
        self.assertFalse(".NAME" in self.element)
        self.element.name = "TestName"
        del self.element.name
        self.assertFalse(".NAME" in self.element)

    def test_data_view(self):
        self.element['NAME'] = "TestName"
        self.assertEqual(self.element.data, {'NAME': 'TestName'})



import unittest

import spydrnet as sdn
from spydrnet.ir.element import Element


class TestNetlist(unittest.TestCase):
    def setUp(self):
        self.netlist = sdn.Netlist()

    def test_constructor(self):
        self.assertIsInstance(self.netlist, Element, "Netlist is not an element.")
        self.assertTrue(self.netlist, "Constructor return None type or empty collection")
        netlist2 = sdn.Netlist()
        self.assertNotEqual(self.netlist, netlist2, "Unique objects are considered equal.")

    def test_libraries(self):
        self.assertEqual(len(tuple(self.netlist.libraries)), 0)
        library = self.netlist.create_library()
        self.assertTrue(self.netlist.libraries[0] == library)
        visited = False
        for visited_library in self.netlist.libraries:
            visited = True
            self.assertEqual(library, visited_library)
        self.assertTrue(visited)

    def test_libraries_set(self):
        library1 = self.netlist.create_library()
        library2 = self.netlist.create_library()
        libraries = [library1, library2]
        self.assertEqual(self.netlist.libraries, libraries)
        self.netlist.libraries = reversed(libraries)
        self.assertEqual(self.netlist.libraries, list(reversed(libraries)))

    def test_top_instance(self):
        self.assertIsNone(self.netlist.top_instance)
        instance = sdn.Instance()
        self.netlist.top_instance = instance
        self.assertEqual(instance, self.netlist.top_instance)
        self.netlist.top_instance = None
        self.assertIsNone(self.netlist.top_instance)

    def test_create_library(self):
        library = self.netlist.create_library()
        self.assertTrue(library in self.netlist.libraries)
        self.assertEqual(library.netlist, self.netlist)

    def test_remove_library(self):
        library = self.netlist.create_library()
        self.netlist.remove_library(library)
        self.assertFalse(library in self.netlist.libraries)
        self.assertIsNone(library.netlist)

    def test_add_library(self):
        library = sdn.Library()
        self.netlist.add_library(library, position=0)
        self.assertTrue(library in self.netlist.libraries)
        self.assertEqual(library.netlist, self.netlist)
        self.assertEqual(list(self.netlist.libraries).count(library), 1)

    @unittest.expectedFailure
    def test_remove_libraries_from_outside_netlist(self):
        library1 = self.netlist.create_library()
        library2 = sdn.Library()
        self.netlist.remove_libraries_from([library1, library2])

    def test_remove_libraries_from(self):
        library = self.netlist.create_library()
        self.netlist.remove_libraries_from((library,))
        self.assertFalse(library in self.netlist)
        self.assertIsNone(library.netlist)
        library_included = self.netlist.create_library()
        library = self.netlist.create_library()
        self.netlist.remove_libraries_from({library})
        self.assertFalse(library in self.netlist)
        self.assertIsNone(library.netlist)
        self.assertTrue(library_included in self.netlist.libraries)
        self.assertEqual(library_included.netlist, self.netlist)


import unittest

import spydrnet as sdn


class TestCable(unittest.TestCase):
    def setUp(self) -> None:
        self.cable = sdn.Cable()

    def test__items(self):
        self.assertEqual(self.cable._items(), self.cable._wires)

    def test_initialize_wires(self):
        self.cable.create_wires(2)
        self.assertEqual(len(self.cable.wires), 2)
        self.cable.remove_wires_from(self.cable.wires)
        self.assertEqual(len(self.cable.wires), 0)

    def test_wires_set(self):
        wire1 = self.cable.create_wire()
        wire2 = self.cable.create_wire()
        self.assertEqual(self.cable.wires, [wire1, wire2])
        self.cable.wires = [wire2, wire1]
        self.assertEqual(self.cable.wires, [wire2, wire1])

    def test_create_wire(self):
        wire = self.cable.create_wire()
        self.assertTrue(wire in self.cable.wires)
        self.assertEqual(wire.cable, self.cable)

    def test_add_wire(self):
        wire = sdn.Wire()
        self.cable.add_wire(wire, position=0)
        self.assertTrue(wire in self.cable.wires)
        self.assertEqual(wire.cable, self.cable)

    def test_remove_wire(self):
        wire = self.cable.create_wire()
        self.cable.remove_wire(wire)
        self.assertFalse(wire in self.cable.wires)
        self.assertIsNone(wire.cable)

    def test_remove_wire_from(self):
        wire_included = self.cable.create_wire()
        wire = self.cable.create_wire()
        self.cable.remove_wires_from({wire})
        self.assertFalse(wire in self.cable.wires)
        self.assertIsNone(wire.cable)
        self.assertTrue(wire_included in self.cable.wires)
        self.assertEqual(wire_included.cable, self.cable)


"""
SpyDrNet
========

SpyDrNet is an EDA tool for netlist analysis and transformation.

See documentation for more details.
"""

# Release data
from spydrnet import release

__author__ = '%s <%s>\n%s <%s>\n%s <%s>' % \
    (release.authors['Keller'] + release.authors['Skouson'] +
        release.authors['Wirthlin'])
__license__ = release.license

__date__ = release.date
__version__ = release.version

from spydrnet.ir import *

OUT = Port.Direction.OUT
IN = Port.Direction.IN
INOUT = Port.Direction.INOUT
UNDEFINED = Port.Direction.UNDEFINED

from spydrnet.testing.test import run as test
from spydrnet.parsers import parse
from spydrnet.composers import compose

import os
base_dir = os.path.dirname(os.path.abspath(__file__))

import glob
example_netlist_names = list()
for filename in glob.glob(os.path.join(base_dir, 'support_files', 'EDIF_netlists', "*")):
    basename = os.path.basename(filename)
    example_netlist_names.append(basename[:basename.index('.')])
example_netlist_names.sort()


def load_example_netlist_by_name(name):
    assert name in example_netlist_names, "Example netlist not found"
    return parse(os.path.join(base_dir, 'support_files', 'EDIF_netlists', name + ".edf.zip"))



import sys
import setuptools
import glob
import os

if sys.argv[-1] == "setup.py":
    print("To install, run 'python setup.py install'")
    print()

if sys.version_info[:2] < (3, 5):
    python_version = "{}.{}".format(sys.version_info[0], sys.version_info[1])
    msg = (
        "SpyDrNet 1.0+ requires Python 3.5 or later ({} detected).\n\n".format(python_version)
    )
    sys.stderr.write(msg + "\n")
    sys.exit(1)

# Write the version information.
sys.path.insert(0, 'spydrnet')
import release
version = release.update_versionfile()
sys.path.pop(0)

with open("README.rst", "r") as fh:
    long_description = fh.read()

example_edif_files = list()
folder_path = os.path.normpath(os.path.join(os.path.dirname(__file__), "spydrnet", "support_files"))
for filename in glob.glob(os.path.join(folder_path, "**", "*"), recursive=True):
    if os.path.isfile(filename) and os.path.getsize(filename) < 1024 * 10:
        example_edif_files.append("support_files/" + str(filename)[len(folder_path) + 1:].replace('\\', '/'))
        
extras_require = {
    "all": [
        "pytest",
    ],
    "pytest": ["pytest"],
}

if __name__ == "__main__":

    setuptools.setup(
        name=release.name.lower(),
        version=version,
        maintainer=release.maintainer,
        maintainer_email=release.maintainer_email,
        author=release.authors['Keller'][0],
        author_email=release.authors['Keller'][1],
        description=release.description,
        keywords=release.keywords,
        long_description=long_description,
        license=release.license,
        platforms=release.platforms,
        url=release.url,
        project_urls=release.project_urls,
        classifiers=release.classifiers,
        package_data={ 'spydrnet': ['VERSION'] + example_edif_files},
        packages=setuptools.find_packages(),
        extras_require=extras_require,
        python_requires='>=3.5',
        zip_safe=False
    )